# Copyright 2014-2021 Canonical Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# This file is sourced from lp:openstack-charm-helpers
#
# Authors:
#  James Page <james.page@ubuntu.com>
#  Adam Gandelman <adamg@ubuntu.com>
#

import collections
import errno
import hashlib
import math
import six

import os
import shutil
import json
import time
import uuid

from subprocess import (
    check_call,
    check_output,
    CalledProcessError,
)
from charmhelpers import deprecate
from charmhelpers.core.hookenv import (
    application_name,
    config,
    service_name,
    local_unit,
    relation_get,
    relation_ids,
    relation_set,
    related_units,
    log,
    DEBUG,
    INFO,
    WARNING,
    ERROR,
)
from charmhelpers.core.host import (
    mount,
    mounts,
    service_start,
    service_stop,
    service_running,
    umount,
    cmp_pkgrevno,
)
from charmhelpers.fetch import (
    apt_install,
)
from charmhelpers.core.unitdata import kv

from charmhelpers.core.kernel import modprobe
from charmhelpers.contrib.openstack.utils import config_flags_parser

KEYRING = '/etc/ceph/ceph.client.{}.keyring'
KEYFILE = '/etc/ceph/ceph.client.{}.key'

CEPH_CONF = """[global]
auth supported = {auth}
keyring = {keyring}
mon host = {mon_hosts}
log to syslog = {use_syslog}
err to syslog = {use_syslog}
clog to syslog = {use_syslog}
"""

# The number of placement groups per OSD to target for placement group
# calculations. This number is chosen as 100 due to the ceph PG Calc
# documentation recommending to choose 100 for clusters which are not
# expected to increase in the foreseeable future. Since the majority of the
# calculations are done on deployment, target the case of non-expanding
# clusters as the default.
DEFAULT_PGS_PER_OSD_TARGET = 100
DEFAULT_POOL_WEIGHT = 10.0
LEGACY_PG_COUNT = 200
DEFAULT_MINIMUM_PGS = 2
AUTOSCALER_DEFAULT_PGS = 32


class OsdPostUpgradeError(Exception):
    """Error class for OSD post-upgrade operations."""
    pass


class OSDSettingConflict(Exception):
    """Error class for conflicting osd setting requests."""
    pass


class OSDSettingNotAllowed(Exception):
    """Error class for a disallowed setting."""
    pass


OSD_SETTING_EXCEPTIONS = (OSDSettingConflict, OSDSettingNotAllowed)

OSD_SETTING_WHITELIST = [
    'osd heartbeat grace',
    'osd heartbeat interval',
]


def _order_dict_by_key(rdict):
    """Convert a dictionary into an OrderedDict sorted by key.

    :param rdict: Dictionary to be ordered.
    :type rdict: dict
    :returns: Ordered Dictionary.
    :rtype: collections.OrderedDict
    """
    return collections.OrderedDict(sorted(rdict.items(), key=lambda k: k[0]))


def get_osd_settings(relation_name):
    """Consolidate requested osd settings from all clients.

    Consolidate requested osd settings from all clients. Check that the
    requested setting is on the whitelist and it does not conflict with
    any other requested settings.

    :returns: Dictionary of settings
    :rtype: dict

    :raises: OSDSettingNotAllowed
    :raises: OSDSettingConflict
    """
    rel_ids = relation_ids(relation_name)
    osd_settings = {}
    for relid in rel_ids:
        for unit in related_units(relid):
            unit_settings = relation_get('osd-settings', unit, relid) or '{}'
            unit_settings = json.loads(unit_settings)
            for key, value in unit_settings.items():
                if key not in OSD_SETTING_WHITELIST:
                    msg = 'Illegal settings "{}"'.format(key)
                    raise OSDSettingNotAllowed(msg)
                if key in osd_settings:
                    if osd_settings[key] != unit_settings[key]:
                        msg = 'Conflicting settings for "{}"'.format(key)
                        raise OSDSettingConflict(msg)
                else:
                    osd_settings[key] = value
    return _order_dict_by_key(osd_settings)


def send_application_name(relid=None):
    """Send the application name down the relation.

    :param relid: Relation id to set application name in.
    :type relid: str
    """
    relation_set(
        relation_id=relid,
        relation_settings={'application-name': application_name()})


def send_osd_settings():
    """Pass on requested OSD settings to osd units."""
    try:
        settings = get_osd_settings('client')
    except OSD_SETTING_EXCEPTIONS as e:
        # There is a problem with the settings, not passing them on. Update
        # status will notify the user.
        log(e, level=ERROR)
        return
    data = {
        'osd-settings': json.dumps(settings, sort_keys=True)}
    for relid in relation_ids('osd'):
        relation_set(relation_id=relid,
                     relation_settings=data)


def validator(value, valid_type, valid_range=None):
    """Helper function for type validation.

    Used to validate these:
    https://docs.ceph.com/docs/master/rados/operations/pools/#set-pool-values
    https://docs.ceph.com/docs/master/rados/configuration/bluestore-config-ref/#inline-compression

    Example input:
        validator(value=1,
                  valid_type=int,
                  valid_range=[0, 2])

    This says I'm testing value=1.  It must be an int inclusive in [0,2]

    :param value: The value to validate.
    :type value: any
    :param valid_type: The type that value should be.
    :type valid_type: any
    :param valid_range: A range of values that value can assume.
    :type valid_range: Optional[Union[List,Tuple]]
    :raises: AssertionError, ValueError
    """
    assert isinstance(value, valid_type), (
        "{} is not a {}".format(value, valid_type))
    if valid_range is not None:
        assert isinstance(
            valid_range, list) or isinstance(valid_range, tuple), (
                "valid_range must be of type List or Tuple, "
                "was given {} of type {}"
                .format(valid_range, type(valid_range)))
        # If we're dealing with strings
        if isinstance(value, six.string_types):
            assert value in valid_range, (
                "{} is not in the list {}".format(value, valid_range))
        # Integer, float should have a min and max
        else:
            if len(valid_range) != 2:
                raise ValueError(
                    "Invalid valid_range list of {} for {}. "
                    "List must be [min,max]".format(valid_range, value))
            assert value >= valid_range[0], (
                "{} is less than minimum allowed value of {}"
                .format(value, valid_range[0]))
            assert value <= valid_range[1], (
                "{} is greater than maximum allowed value of {}"
                .format(value, valid_range[1]))


class PoolCreationError(Exception):
    """A custom exception to inform the caller that a pool creation failed.

    Provides an error message
    """

    def __init__(self, message):
        super(PoolCreationError, self).__init__(message)


class BasePool(object):
    """An object oriented approach to Ceph pool creation.

    This base class is inherited by ReplicatedPool and ErasurePool. Do not call
    create() on this base class as it will raise an exception.

    Instantiate a child class and call create().
    """
    # Dictionary that maps pool operation properties to Tuples with valid type
    # and valid range
    op_validation_map = {
        'compression-algorithm': (str, ('lz4', 'snappy', 'zlib', 'zstd')),
        'compression-mode': (str, ('none', 'passive', 'aggressive', 'force')),
        'compression-required-ratio': (float, None),
        'compression-min-blob-size': (int, None),
        'compression-min-blob-size-hdd': (int, None),
        'compression-min-blob-size-ssd': (int, None),
        'compression-max-blob-size': (int, None),
        'compression-max-blob-size-hdd': (int, None),
        'compression-max-blob-size-ssd': (int, None),
        'rbd-mirroring-mode': (str, ('image', 'pool'))
    }

    def __init__(self, service, name=None, percent_data=None, app_name=None,
                 op=None):
        """Initialize BasePool object.

        Pool information is either initialized from individual keyword
        arguments or from a individual CephBrokerRq operation Dict.

        :param service: The Ceph user name to run commands under.
        :type service: str
        :param name: Name of pool to operate on.
        :type name: str
        :param percent_data: The expected pool size in relation to all
                             available resources in the Ceph cluster. Will be
                             used to set the ``target_size_ratio`` pool
                             property. (default: 10.0)
        :type percent_data: Optional[float]
        :param app_name: Ceph application name, usually one of:
                         ('cephfs', 'rbd', 'rgw') (default: 'unknown')
        :type app_name: Optional[str]
        :param op: Broker request Op to compile pool data from.
        :type op: Optional[Dict[str,any]]
        :raises: KeyError
        """
        # NOTE: Do not perform initialization steps that require live data from
        # a running cluster here. The *Pool classes may be used for validation.
        self.service = service
        self.op = op or {}

        if op:
            # When initializing from op the `name` attribute is required and we
            # will fail with KeyError if it is not provided.
            self.name = op['name']
            self.percent_data = op.get('weight')
            self.app_name = op.get('app-name')
        else:
            self.name = name
            self.percent_data = percent_data
            self.app_name = app_name

        # Set defaults for these if they are not provided
        self.percent_data = self.percent_data or 10.0
        self.app_name = self.app_name or 'unknown'

    def validate(self):
        """Check that value of supplied operation parameters are valid.

        :raises: ValueError
        """
        for op_key, op_value in self.op.items():
            if op_key in self.op_validation_map and op_value is not None:
                valid_type, valid_range = self.op_validation_map[op_key]
                try:
                    validator(op_value, valid_type, valid_range)
                except (AssertionError, ValueError) as e:
                    # Normalize on ValueError, also add information about which
                    # variable we had an issue with.
                    raise ValueError("'{}': {}".format(op_key, str(e)))

    def _create(self):
        """Perform the pool creation, method MUST be overridden by child class.
        """
        raise NotImplementedError

    def _post_create(self):
        """Perform common post pool creation tasks.

        Note that pool properties subject to change during the lifetime of a
        pool / deployment should go into the ``update`` method.

        Do not add calls for a specific pool type here, those should go into
        one of the pool specific classes.
        """
        nautilus_or_later = cmp_pkgrevno('ceph-common', '14.2.0') >= 0
        if nautilus_or_later:
            # Ensure we set the expected pool ratio
            update_pool(
                client=self.service,
                pool=self.name,
                settings={
                    'target_size_ratio': str(
                        self.percent_data / 100.0),
                })
        try:
            set_app_name_for_pool(client=self.service,
                                  pool=self.name,
                                  name=self.app_name)
        except CalledProcessError:
            log('Could not set app name for pool {}'
                .format(self.name),
                level=WARNING)
        if 'pg_autoscaler' in enabled_manager_modules():
            try:
                enable_pg_autoscale(self.service, self.name)
            except CalledProcessError as e:
                log('Could not configure auto scaling for pool {}: {}'
                    .format(self.name, e),
                    level=WARNING)

    def create(self):
        """Create pool and perform any post pool creation tasks.

        To allow for sharing of common code among pool specific classes the
        processing has been broken out into the private methods ``_create``
        and ``_post_create``.

        Do not add any pool type specific handling here, that should go into
        one of the pool specific classes.
        """
        if not pool_exists(self.service, self.name):
            self.validate()
            self._create()
            self._post_create()
            self.update()

    def set_quota(self):
        """Set a quota if requested.

        :raises: CalledProcessError
        """
        max_bytes = self.op.get('max-bytes')
        max_objects = self.op.get('max-objects')
        if max_bytes or max_objects:
            set_pool_quota(service=self.service, pool_name=self.name,
                           max_bytes=max_bytes, max_objects=max_objects)

    def set_compression(self):
        """Set compression properties if requested.

        :raises: CalledProcessError
        """
        compression_properties = {
            key.replace('-', '_'): value
            for key, value in self.op.items()
            if key in (
                'compression-algorithm',
                'compression-mode',
                'compression-required-ratio',
                'compression-min-blob-size',
                'compression-min-blob-size-hdd',
                'compression-min-blob-size-ssd',
                'compression-max-blob-size',
                'compression-max-blob-size-hdd',
                'compression-max-blob-size-ssd') and value}
        if compression_properties:
            update_pool(self.service, self.name, compression_properties)

    def update(self):
        """Update properties for an already existing pool.

        Do not add calls for a specific pool type here, those should go into
        one of the pool specific classes.
        """
        self.validate()
        self.set_quota()
        self.set_compression()

    def add_cache_tier(self, cache_pool, mode):
        """Adds a new cache tier to an existing pool.

        :param cache_pool: The cache tier pool name to add.
        :type cache_pool: str
        :param mode: The caching mode to use for this pool.
                     valid range = ["readonly", "writeback"]
        :type mode: str
        """
        # Check the input types and values
        validator(value=cache_pool, valid_type=six.string_types)
        validator(
            value=mode, valid_type=six.string_types,
            valid_range=["readonly", "writeback"])

        check_call([
            'ceph', '--id', self.service,
            'osd', 'tier', 'add', self.name, cache_pool,
        ])
        check_call([
            'ceph', '--id', self.service,
            'osd', 'tier', 'cache-mode', cache_pool, mode,
        ])
        check_call([
            'ceph', '--id', self.service,
            'osd', 'tier', 'set-overlay', self.name, cache_pool,
        ])
        check_call([
            'ceph', '--id', self.service,
            'osd', 'pool', 'set', cache_pool, 'hit_set_type', 'bloom',
        ])

    def remove_cache_tier(self, cache_pool):
        """Removes a cache tier from Ceph.

        Flushes all dirty objects from writeback pools and waits for that to
        complete.

        :param cache_pool: The cache tier pool name to remove.
        :type cache_pool: str
        """
        # read-only is easy, writeback is much harder
        mode = get_cache_mode(self.service, cache_pool)
        if mode == 'readonly':
            check_call([
                'ceph', '--id', self.service,
                'osd', 'tier', 'cache-mode', cache_pool, 'none'
            ])
            check_call([
                'ceph', '--id', self.service,
                'osd', 'tier', 'remove', self.name, cache_pool,
            ])

        elif mode == 'writeback':
            pool_forward_cmd = ['ceph', '--id', self.service, 'osd', 'tier',
                                'cache-mode', cache_pool, 'forward']
            if cmp_pkgrevno('ceph-common', '10.1') >= 0:
                # Jewel added a mandatory flag
                pool_forward_cmd.append('--yes-i-really-mean-it')

            check_call(pool_forward_cmd)
            # Flush the cache and wait for it to return
            check_call([
                'rados', '--id', self.service,
                '-p', cache_pool, 'cache-flush-evict-all'])
            check_call([
                'ceph', '--id', self.service,
                'osd', 'tier', 'remove-overlay', self.name])
            check_call([
                'ceph', '--id', self.service,
                'osd', 'tier', 'remove', self.name, cache_pool])

    def get_pgs(self, pool_size, percent_data=DEFAULT_POOL_WEIGHT,
                device_class=None):
        """Return the number of placement groups to use when creating the pool.

        Returns the number of placement groups which should be specified when
        creating the pool. This is based upon the calculation guidelines
        provided by the Ceph Placement Group Calculator (located online at
        http://ceph.com/pgcalc/).

        The number of placement groups are calculated using the following:

            (Target PGs per OSD) * (OSD #) * (%Data)
            ----------------------------------------
                         (Pool size)

        Per the upstream guidelines, the OSD # should really be considered
        based on the number of OSDs which are eligible to be selected by the
        pool. Since the pool creation doesn't specify any of CRUSH set rules,
        the default rule will be dependent upon the type of pool being
        created (replicated or erasure).

        This code makes no attempt to determine the number of OSDs which can be
        selected for the specific rule, rather it is left to the user to tune
        in the form of 'expected-osd-count' config option.

        :param pool_size: pool_size is either the number of replicas for
            replicated pools or the K+M sum for erasure coded pools
        :type pool_size: int
        :param percent_data: the percentage of data that is expected to
            be contained in the pool for the specific OSD set. Default value
            is to assume 10% of the data is for this pool, which is a
            relatively low % of the data but allows for the pg_num to be
            increased. NOTE: the default is primarily to handle the scenario
            where related charms requiring pools has not been upgraded to
            include an update to indicate their relative usage of the pools.
        :type percent_data: float
        :param device_class: class of storage to use for basis of pgs
            calculation; ceph supports nvme, ssd and hdd by default based
            on presence of devices of each type in the deployment.
        :type device_class: str
        :returns: The number of pgs to use.
        :rtype: int
        """

        # Note: This calculation follows the approach that is provided
        # by the Ceph PG Calculator located at http://ceph.com/pgcalc/.
        validator(value=pool_size, valid_type=int)

        # Ensure that percent data is set to something - even with a default
        # it can be set to None, which would wreak havoc below.
        if percent_data is None:
            percent_data = DEFAULT_POOL_WEIGHT

        # If the expected-osd-count is specified, then use the max between
        # the expected-osd-count and the actual osd_count
        osd_list = get_osds(self.service, device_class)
        expected = config('expected-osd-count') or 0

        if osd_list:
            if device_class:
                osd_count = len(osd_list)
            else:
                osd_count = max(expected, len(osd_list))

            # Log a message to provide some insight if the calculations claim
            # to be off because someone is setting the expected count and
            # there are more OSDs in reality. Try to make a proper guess
            # based upon the cluster itself.
            if not device_class and expected and osd_count != expected:
                log("Found more OSDs than provided expected count. "
                    "Using the actual count instead", INFO)
        elif expected:
            # Use the expected-osd-count in older ceph versions to allow for
            # a more accurate pg calculations
            osd_count = expected
        else:
            # NOTE(james-page): Default to 200 for older ceph versions
            # which don't support OSD query from cli
            return LEGACY_PG_COUNT

        percent_data /= 100.0
        target_pgs_per_osd = config(
            'pgs-per-osd') or DEFAULT_PGS_PER_OSD_TARGET
        num_pg = (target_pgs_per_osd * osd_count * percent_data) // pool_size

        # NOTE: ensure a sane minimum number of PGS otherwise we don't get any
        #       reasonable data distribution in minimal OSD configurations
        if num_pg < DEFAULT_MINIMUM_PGS:
            num_pg = DEFAULT_MINIMUM_PGS

        # The CRUSH algorithm has a slight optimization for placement groups
        # with powers of 2 so find the nearest power of 2. If the nearest
        # power of 2 is more than 25% below the original value, the next
        # highest value is used. To do this, find the nearest power of 2 such
        # that 2^n <= num_pg, check to see if its within the 25% tolerance.
        exponent = math.floor(math.log(num_pg, 2))
        nearest = 2 ** exponent
        if (num_pg - nearest) > (num_pg * 0.25):
            # Choose the next highest power of 2 since the nearest is more
            # than 25% below the original value.
            return int(nearest * 2)
        else:
            return int(nearest)


class Pool(BasePool):
    """Compatibility shim for any descendents external to this library."""

    @deprecate(
        'The ``Pool`` baseclass has been replaced by ``BasePool`` class.')
    def __init__(self, service, name):
        super(Pool, self).__init__(service, name=name)

    def create(self):
        pass


class ReplicatedPool(BasePool):
    def __init__(self, service, name=None, pg_num=None, replicas=None,
                 percent_data=None, app_name=None, op=None):
        """Initialize ReplicatedPool object.

        Pool information is either initialized from individual keyword
        arguments or from a individual CephBrokerRq operation Dict.

        Please refer to the docstring of the ``BasePool`` class for
        documentation of the common parameters.

        :param pg_num: Express wish for number of Placement Groups (this value
                       is subject to validation against a running cluster prior
                       to use to avoid creating a pool with too many PGs)
        :type pg_num: int
        :param replicas: Number of copies there should be of each object added
                         to this replicated pool.
        :type replicas: int
        :raises: KeyError
        """
        # NOTE: Do not perform initialization steps that require live data from
        # a running cluster here. The *Pool classes may be used for validation.

        # The common parameters are handled in our parents initializer
        super(ReplicatedPool, self).__init__(
            service=service, name=name, percent_data=percent_data,
            app_name=app_name, op=op)

        if op:
            # When initializing from op `replicas` is a required attribute, and
            # we will fail with KeyError if it is not provided.
            self.replicas = op['replicas']
            self.pg_num = op.get('pg_num')
        else:
            self.replicas = replicas or 2
            self.pg_num = pg_num

    def _create(self):
        # Do extra validation on pg_num with data from live cluster
        if self.pg_num:
            # Since the number of placement groups were specified, ensure
            # that there aren't too many created.
            max_pgs = self.get_pgs(self.replicas, 100.0)
            self.pg_num = min(self.pg_num, max_pgs)
        else:
            self.pg_num = self.get_pgs(self.replicas, self.percent_data)

        nautilus_or_later = cmp_pkgrevno('ceph-common', '14.2.0') >= 0
        # Create it
        if nautilus_or_later:
            cmd = [
                'ceph', '--id', self.service, 'osd', 'pool', 'create',
                '--pg-num-min={}'.format(
                    min(AUTOSCALER_DEFAULT_PGS, self.pg_num)
                ),
                self.name, str(self.pg_num)
            ]
        else:
            cmd = [
                'ceph', '--id', self.service, 'osd', 'pool', 'create',
                self.name, str(self.pg_num)
            ]
        check_call(cmd)

    def _post_create(self):
        # Set the pool replica size
        update_pool(client=self.service,
                    pool=self.name,
                    settings={'size': str(self.replicas)})
        # Perform other common post pool creation tasks
        super(ReplicatedPool, self)._post_create()


class ErasurePool(BasePool):
    """Default jerasure erasure coded pool."""

    def __init__(self, service, name=None, erasure_code_profile=None,
                 percent_data=None, app_name=None, op=None,
                 allow_ec_overwrites=False):
        """Initialize ReplicatedPool object.

        Pool information is either initialized from individual keyword
        arguments or from a individual CephBrokerRq operation Dict.

        Please refer to the docstring of the ``BasePool`` class for
        documentation of the common parameters.

        :param erasure_code_profile: EC Profile to use (default: 'default')
        :type erasure_code_profile: Optional[str]
        """
        # NOTE: Do not perform initialization steps that require live data from
        # a running cluster here. The *Pool classes may be used for validation.

        # The common parameters are handled in our parents initializer
        super(ErasurePool, self).__init__(
            service=service, name=name, percent_data=percent_data,
            app_name=app_name, op=op)

        if op:
            # Note that the different default when initializing from op stems
            # from different handling of this in the `charms.ceph` library.
            self.erasure_code_profile = op.get('erasure-profile',
                                               'default-canonical')
            self.allow_ec_overwrites = op.get('allow-ec-overwrites')
        else:
            # We keep the class default when initialized from keyword arguments
            # to not break the API for any other consumers.
            self.erasure_code_profile = erasure_code_profile or 'default'
            self.allow_ec_overwrites = allow_ec_overwrites

    def _create(self):
        # Try to find the erasure profile information in order to properly
        # size the number of placement groups. The size of an erasure
        # coded placement group is calculated as k+m.
        erasure_profile = get_erasure_profile(self.service,
                                              self.erasure_code_profile)

        # Check for errors
        if erasure_profile is None:
            msg = ("Failed to discover erasure profile named "
                   "{}".format(self.erasure_code_profile))
            log(msg, level=ERROR)
            raise PoolCreationError(msg)
        if 'k' not in erasure_profile or 'm' not in erasure_profile:
            # Error
            msg = ("Unable to find k (data chunks) or m (coding chunks) "
                   "in erasure profile {}".format(erasure_profile))
            log(msg, level=ERROR)
            raise PoolCreationError(msg)

        k = int(erasure_profile['k'])
        m = int(erasure_profile['m'])
        pgs = self.get_pgs(k + m, self.percent_data)
        nautilus_or_later = cmp_pkgrevno('ceph-common', '14.2.0') >= 0
        # Create it
        if nautilus_or_later:
            cmd = [
                'ceph', '--id', self.service, 'osd', 'pool', 'create',
                '--pg-num-min={}'.format(
                    min(AUTOSCALER_DEFAULT_PGS, pgs)
                ),
                self.name, str(pgs), str(pgs),
                'erasure', self.erasure_code_profile
            ]
        else:
            cmd = [
                'ceph', '--id', self.service, 'osd', 'pool', 'create',
                self.name, str(pgs), str(pgs),
                'erasure', self.erasure_code_profile
            ]
        check_call(cmd)

    def _post_create(self):
        super(ErasurePool, self)._post_create()
        if self.allow_ec_overwrites:
            update_pool(self.service, self.name,
                        {'allow_ec_overwrites': 'true'})


def enabled_manager_modules():
    """Return a list of enabled manager modules.

    :rtype: List[str]
    """
    cmd = ['ceph', 'mgr', 'module', 'ls']
    try:
        modules = check_output(cmd)
        if six.PY3:
            modules = modules.decode('UTF-8')
    except CalledProcessError as e:
        log("Failed to list ceph modules: {}".format(e), WARNING)
        return []
    modules = json.loads(modules)
    return modules['enabled_modules']


def enable_pg_autoscale(service, pool_name):
    """Enable Ceph's PG autoscaler for the specified pool.

    :param service: The Ceph user name to run the command under
    :type service: str
    :param pool_name: The name of the pool to enable sutoscaling on
    :type pool_name: str
    :raises: CalledProcessError if the command fails
    """
    check_call([
        'ceph', '--id', service,
        'osd', 'pool', 'set', pool_name, 'pg_autoscale_mode', 'on'])


def get_mon_map(service):
    """Return the current monitor map.

    :param service: The Ceph user name to run the command under
    :type service: str
    :returns: Dictionary with monitor map data
    :rtype: Dict[str,any]
    :raises: ValueError if the monmap fails to parse, CalledProcessError if our
             ceph command fails.
    """
    try:
        octopus_or_later = cmp_pkgrevno('ceph-common', '15.0.0') >= 0
        mon_status_cmd = 'quorum_status' if octopus_or_later else 'mon_status'
        mon_status = check_output(['ceph', '--id', service,
                                   mon_status_cmd, '--format=json'])
        if six.PY3:
            mon_status = mon_status.decode('UTF-8')
        try:
            return json.loads(mon_status)
        except ValueError as v:
            log("Unable to parse mon_status json: {}. Error: {}"
                .format(mon_status, str(v)))
            raise
    except CalledProcessError as e:
        log("mon_status command failed with message: {}"
            .format(str(e)))
        raise


def hash_monitor_names(service):
    """Get a sorted list of monitor hashes in ascending order.

    Uses the get_mon_map() function to get information about the monitor
    cluster. Hash the name of each monitor.

    :param service: The Ceph user name to run the command under.
    :type service: str
    :returns: a sorted list of monitor hashes in an ascending order.
    :rtype : List[str]
    :raises: CalledProcessError, ValueError
    """
    try:
        hash_list = []
        monitor_list = get_mon_map(service=service)
        if monitor_list['monmap']['mons']:
            for mon in monitor_list['monmap']['mons']:
                hash_list.append(
                    hashlib.sha224(mon['name'].encode('utf-8')).hexdigest())
            return sorted(hash_list)
        else:
            return None
    except (ValueError, CalledProcessError):
        raise


def monitor_key_delete(service, key):
    """Delete a key and value pair from the monitor cluster.

    Deletes a key value pair on the monitor cluster.

    :param service: The Ceph user name to run the command under
    :type service: str
    :param key: The key to delete.
    :type key: str
    :raises: CalledProcessError
    """
    try:
        check_output(
            ['ceph', '--id', service,
             'config-key', 'del', str(key)])
    except CalledProcessError as e:
        log("Monitor config-key put failed with message: {}"
            .format(e.output))
        raise


def monitor_key_set(service, key, value):
    """Set a key value pair on the monitor cluster.

    :param service: The Ceph user name to run the command under.
    :type service str
    :param key: The key to set.
    :type key: str
    :param value: The value to set. This will be coerced into a string.
    :type value: str
    :raises: CalledProcessError
    """
    try:
        check_output(
            ['ceph', '--id', service,
             'config-key', 'put', str(key), str(value)])
    except CalledProcessError as e:
        log("Monitor config-key put failed with message: {}"
            .format(e.output))
        raise


def monitor_key_get(service, key):
    """Get the value of an existing key in the monitor cluster.

    :param service: The Ceph user name to run the command under
    :type service: str
    :param key: The key to search for.
    :type key: str
    :return: Returns the value of that key or None if not found.
    :rtype: Optional[str]
    """
    try:
        output = check_output(
            ['ceph', '--id', service,
             'config-key', 'get', str(key)]).decode('UTF-8')
        return output
    except CalledProcessError as e:
        log("Monitor config-key get failed with message: {}"
            .format(e.output))
        return None


def monitor_key_exists(service, key):
    """Search for existence of key in the monitor cluster.

    :param service: The Ceph user name to run the command under.
    :type service: str
    :param key: The key to search for.
    :type key: str
    :return: Returns True if the key exists, False if not.
    :rtype: bool
    :raises: CalledProcessError if an unknown error occurs.
    """
    try:
        check_call(
            ['ceph', '--id', service,
             'config-key', 'exists', str(key)])
        # I can return true here regardless because Ceph returns
        # ENOENT if the key wasn't found
        return True
    except CalledProcessError as e:
        if e.returncode == errno.ENOENT:
            return False
        else:
            log("Unknown error from ceph config-get exists: {} {}"
                .format(e.returncode, e.output))
            raise


def get_erasure_profile(service, name):
    """Get an existing erasure code profile if it exists.

    :param service: The Ceph user name to run the command under.
    :type service: str
    :param name: Name of profile.
    :type name: str
    :returns: Dictionary with profile data.
    :rtype: Optional[Dict[str]]
    """
    try:
        out = check_output(['ceph', '--id', service,
                            'osd', 'erasure-code-profile', 'get',
                            name, '--format=json'])
        if six.PY3:
            out = out.decode('UTF-8')
        return json.loads(out)
    except (CalledProcessError, OSError, ValueError):
        return None


def pool_set(service, pool_name, key, value):
    """Sets a value for a RADOS pool in ceph.

    :param service: The Ceph user name to run the command under.
    :type service: str
    :param pool_name: Name of pool to set property on.
    :type pool_name: str
    :param key: Property key.
    :type key: str
    :param value: Value, will be coerced into str and shifted to lowercase.
    :type value: str
    :raises: CalledProcessError
    """
    cmd = [
        'ceph', '--id', service,
        'osd', 'pool', 'set', pool_name, key, str(value).lower()]
    check_call(cmd)


def snapshot_pool(service, pool_name, snapshot_name):
    """Snapshots a RADOS pool in Ceph.

    :param service: The Ceph user name to run the command under.
    :type service: str
    :param pool_name: Name of pool to snapshot.
    :type pool_name: str
    :param snapshot_name: Name of snapshot to create.
    :type snapshot_name: str
    :raises: CalledProcessError
    """
    cmd = [
        'ceph', '--id', service,
        'osd', 'pool', 'mksnap', pool_name, snapshot_name]
    check_call(cmd)


def remove_pool_snapshot(service, pool_name, snapshot_name):
    """Remove a snapshot from a RADOS pool in Ceph.

    :param service: The Ceph user name to run the command under.
    :type service: str
    :param pool_name: Name of pool to remove snapshot from.
    :type pool_name: str
    :param snapshot_name: Name of snapshot to remove.
    :type snapshot_name: str
    :raises: CalledProcessError
    """
    cmd = [
        'ceph', '--id', service,
        'osd', 'pool', 'rmsnap', pool_name, snapshot_name]
    check_call(cmd)


def set_pool_quota(service, pool_name, max_bytes=None, max_objects=None):
    """Set byte quota on a RADOS pool in Ceph.

    :param service: The Ceph user name to run the command under
    :type service: str
    :param pool_name: Name of pool
    :type pool_name: str
    :param max_bytes: Maximum bytes quota to apply
    :type max_bytes: int
    :param max_objects: Maximum objects quota to apply
    :type max_objects: int
    :raises: subprocess.CalledProcessError
    """
    cmd = [
        'ceph', '--id', service,
        'osd', 'pool', 'set-quota', pool_name]
    if max_bytes:
        cmd = cmd + ['max_bytes', str(max_bytes)]
    if max_objects:
        cmd = cmd + ['max_objects', str(max_objects)]
    check_call(cmd)


def remove_pool_quota(service, pool_name):
    """Remove byte quota on a RADOS pool in Ceph.

    :param service: The Ceph user name to run the command under.
    :type service: str
    :param pool_name: Name of pool to remove quota from.
    :type pool_name: str
    :raises: CalledProcessError
    """
    cmd = [
        'ceph', '--id', service,
        'osd', 'pool', 'set-quota', pool_name, 'max_bytes', '0']
    check_call(cmd)


def remove_erasure_profile(service, profile_name):
    """Remove erasure code profile.

    :param service: The Ceph user name to run the command under
    :type service: str
    :param profile_name: Name of profile to remove.
    :type profile_name: str
    :raises: CalledProcessError
    """
    cmd = [
        'ceph', '--id', service,
        'osd', 'erasure-code-profile', 'rm', profile_name]
    check_call(cmd)


def create_erasure_profile(service, profile_name,
                           erasure_plugin_name='jerasure',
                           failure_domain=None,
                           data_chunks=2, coding_chunks=1,
                           locality=None, durability_estimator=None,
                           helper_chunks=None,
                           scalar_mds=None,
                           crush_locality=None,
                           device_class=None,
                           erasure_plugin_technique=None):
    """Create a new erasure code profile if one does not already exist for it.

    Profiles are considered immutable so will not be updated if the named
    profile already exists.

    Please refer to [0] for more details.

    0: http://docs.ceph.com/docs/master/rados/operations/erasure-code-profile/

    :param service: The Ceph user name to run the command under.
    :type service: str
    :param profile_name: Name of profile.
    :type profile_name: str
    :param erasure_plugin_name: Erasure code plugin.
    :type erasure_plugin_name: str
    :param failure_domain: Failure domain, one of:
                           ('chassis', 'datacenter', 'host', 'osd', 'pdu',
                            'pod', 'rack', 'region', 'room', 'root', 'row').
    :type failure_domain: str
    :param data_chunks: Number of data chunks.
    :type data_chunks: int
    :param coding_chunks: Number of coding chunks.
    :type coding_chunks: int
    :param locality: Locality.
    :type locality: int
    :param durability_estimator: Durability estimator.
    :type durability_estimator: int
    :param helper_chunks: int
    :type helper_chunks: int
    :param device_class: Restrict placement to devices of specific class.
    :type device_class: str
    :param scalar_mds: one of ['isa', 'jerasure', 'shec']
    :type scalar_mds: str
    :param crush_locality: LRC locality faulure domain, one of:
                           ('chassis', 'datacenter', 'host', 'osd', 'pdu', 'pod',
                            'rack', 'region', 'room', 'root', 'row') or unset.
    :type crush_locaity: str
    :param erasure_plugin_technique: Coding technique for EC plugin
    :type erasure_plugin_technique: str
    :return: None.  Can raise CalledProcessError, ValueError or AssertionError
    """
    if erasure_profile_exists(service, profile_name):
        log('EC profile {} exists, skipping update'.format(profile_name),
            level=WARNING)
        return

    plugin_techniques = {
        'jerasure': [
            'reed_sol_van',
            'reed_sol_r6_op',
            'cauchy_orig',
            'cauchy_good',
            'liberation',
            'blaum_roth',
            'liber8tion'
        ],
        'lrc': [],
        'isa': [
            'reed_sol_van',
            'cauchy',
        ],
        'shec': [
            'single',
            'multiple'
        ],
        'clay': [],
    }
    failure_domains = [
        'chassis', 'datacenter',
        'host', 'osd',
        'pdu', 'pod',
        'rack', 'region',
        'room', 'root',
        'row',
    ]
    device_classes = [
        'ssd',
        'hdd',
        'nvme'
    ]

    validator(erasure_plugin_name, six.string_types,
              list(plugin_techniques.keys()))

    cmd = [
        'ceph', '--id', service,
        'osd', 'erasure-code-profile', 'set', profile_name,
        'plugin={}'.format(erasure_plugin_name),
        'k={}'.format(str(data_chunks)),
        'm={}'.format(str(coding_chunks)),
    ]

    if erasure_plugin_technique:
        validator(erasure_plugin_technique, six.string_types,
                  plugin_techniques[erasure_plugin_name])
        cmd.append('technique={}'.format(erasure_plugin_technique))

    luminous_or_later = cmp_pkgrevno('ceph-common', '12.0.0') >= 0

    # Set failure domain from options if not provided in args
    if not failure_domain and config('customize-failure-domain'):
        # Defaults to 'host' so just need to deal with
        # setting 'rack' if feature is enabled
        failure_domain = 'rack'

    if failure_domain:
        validator(failure_domain, six.string_types, failure_domains)
        # failure_domain changed in luminous
        if luminous_or_later:
            cmd.append('crush-failure-domain={}'.format(failure_domain))
        else:
            cmd.append('ruleset-failure-domain={}'.format(failure_domain))

    # device class new in luminous
    if luminous_or_later and device_class:
        validator(device_class, six.string_types, device_classes)
        cmd.append('crush-device-class={}'.format(device_class))
    else:
        log('Skipping device class configuration (ceph < 12.0.0)',
            level=DEBUG)

    # Add plugin specific information
    if erasure_plugin_name == 'lrc':
        # LRC mandatory configuration
        if locality:
            cmd.append('l={}'.format(str(locality)))
        else:
            raise ValueError("locality must be provided for lrc plugin")
        # LRC optional configuration
        if crush_locality:
            validator(crush_locality, six.string_types, failure_domains)
            cmd.append('crush-locality={}'.format(crush_locality))

    if erasure_plugin_name == 'shec':
        # SHEC optional configuration
        if durability_estimator:
            cmd.append('c={}'.format((durability_estimator)))

    if erasure_plugin_name == 'clay':
        # CLAY optional configuration
        if helper_chunks:
            cmd.append('d={}'.format(str(helper_chunks)))
        if scalar_mds:
            cmd.append('scalar-mds={}'.format(scalar_mds))

    check_call(cmd)


def rename_pool(service, old_name, new_name):
    """Rename a Ceph pool from old_name to new_name.

    :param service: The Ceph user name to run the command under.
    :type service: str
    :param old_name: Name of pool subject to rename.
    :type old_name: str
    :param new_name: Name to rename pool to.
    :type new_name: str
    """
    validator(value=old_name, valid_type=six.string_types)
    validator(value=new_name, valid_type=six.string_types)

    cmd = [
        'ceph', '--id', service,
        'osd', 'pool', 'rename', old_name, new_name]
    check_call(cmd)


def erasure_profile_exists(service, name):
    """Check to see if an Erasure code profile already exists.

    :param service: The Ceph user name to run the command under
    :type service: str
    :param name: Name of profile to look for.
    :type name: str
    :returns: True if it exists, False otherwise.
    :rtype: bool
    """
    validator(value=name, valid_type=six.string_types)
    try:
        check_call(['ceph', '--id', service,
                    'osd', 'erasure-code-profile', 'get',
                    name])
        return True
    except CalledProcessError:
        return False


def get_cache_mode(service, pool_name):
    """Find the current caching mode of the pool_name given.

    :param service: The Ceph user name to run the command under
    :type service: str
    :param pool_name: Name of pool.
    :type pool_name: str
    :returns: Current cache mode.
    :rtype: Optional[int]
    """
    validator(value=service, valid_type=six.string_types)
    validator(value=pool_name, valid_type=six.string_types)
    out = check_output(['ceph', '--id', service,
                        'osd', 'dump', '--format=json'])
    if six.PY3:
        out = out.decode('UTF-8')
    try:
        osd_json = json.loads(out)
        for pool in osd_json['pools']:
            if pool['pool_name'] == pool_name:
                return pool['cache_mode']
        return None
    except ValueError:
        raise


def pool_exists(service, name):
    """Check to see if a RADOS pool already exists."""
    try:
        out = check_output(['rados', '--id', service, 'lspools'])
        if six.PY3:
            out = out.decode('UTF-8')
    except CalledProcessError:
        return False

    return name in out.split()


def get_osds(service, device_class=None):
    """Return a list of all Ceph Object Storage Daemons currently in the
    cluster (optionally filtered by storage device class).

    :param device_class: Class of storage device for OSD's
    :type device_class: str
    """
    luminous_or_later = cmp_pkgrevno('ceph-common', '12.0.0') >= 0
    if luminous_or_later and device_class:
        out = check_output(['ceph', '--id', service,
                            'osd', 'crush', 'class',
                            'ls-osd', device_class,
                            '--format=json'])
    else:
        out = check_output(['ceph', '--id', service,
                            'osd', 'ls',
                            '--format=json'])
    if six.PY3:
        out = out.decode('UTF-8')
    return json.loads(out)


def install():
    """Basic Ceph client installation."""
    ceph_dir = "/etc/ceph"
    if not os.path.exists(ceph_dir):
        os.mkdir(ceph_dir)

    apt_install('ceph-common', fatal=True)


def rbd_exists(service, pool, rbd_img):
    """Check to see if a RADOS block device exists."""
    try:
        out = check_output(['rbd', 'list', '--id',
                            service, '--pool', pool])
        if six.PY3:
            out = out.decode('UTF-8')
    except CalledProcessError:
        return False

    return rbd_img in out


def create_rbd_image(service, pool, image, sizemb):
    """Create a new RADOS block device."""
    cmd = ['rbd', 'create', image, '--size', str(sizemb), '--id', service,
           '--pool', pool]
    check_call(cmd)


def update_pool(client, pool, settings):
    """Update pool properties.

    :param client: Client/User-name to authenticate with.
    :type client: str
    :param pool: Name of pool to operate on
    :type pool: str
    :param settings: Dictionary with key/value pairs to set.
    :type settings: Dict[str, str]
    :raises: CalledProcessError
    """
    cmd = ['ceph', '--id', client, 'osd', 'pool', 'set', pool]
    for k, v in six.iteritems(settings):
        check_call(cmd + [k, v])


def set_app_name_for_pool(client, pool, name):
    """Calls `osd pool application enable` for the specified pool name

    :param client: Name of the ceph client to use
    :type client: str
    :param pool: Pool to set app name for
    :type pool: str
    :param name: app name for the specified pool
    :type name: str

    :raises: CalledProcessError if ceph call fails
    """
    if cmp_pkgrevno('ceph-common', '12.0.0') >= 0:
        cmd = ['ceph', '--id', client, 'osd', 'pool',
               'application', 'enable', pool, name]
        check_call(cmd)


def create_pool(service, name, replicas=3, pg_num=None):
    """Create a new RADOS pool."""
    if pool_exists(service, name):
        log("Ceph pool {} already exists, skipping creation".format(name),
            level=WARNING)
        return

    if not pg_num:
        # Calculate the number of placement groups based
        # on upstream recommended best practices.
        osds = get_osds(service)
        if osds:
            pg_num = (len(osds) * 100 // replicas)
        else:
            # NOTE(james-page): Default to 200 for older ceph versions
            # which don't support OSD query from cli
            pg_num = 200

    cmd = ['ceph', '--id', service, 'osd', 'pool', 'create', name, str(pg_num)]
    check_call(cmd)

    update_pool(service, name, settings={'size': str(replicas)})


def delete_pool(service, name):
    """Delete a RADOS pool from ceph."""
    cmd = ['ceph', '--id', service, 'osd', 'pool', 'delete', name,
           '--yes-i-really-really-mean-it']
    check_call(cmd)


def _keyfile_path(service):
    return KEYFILE.format(service)


def _keyring_path(service):
    return KEYRING.format(service)


def add_key(service, key):
    """Add a key to a keyring.

    Creates the keyring if it doesn't already exist.

    Logs and returns if the key is already in the keyring.
    """
    keyring = _keyring_path(service)
    if os.path.exists(keyring):
        with open(keyring, 'r') as ring:
            if key in ring.read():
                log('Ceph keyring exists at %s and has not changed.' % keyring,
                    level=DEBUG)
                return
            log('Updating existing keyring %s.' % keyring, level=DEBUG)

    cmd = ['ceph-authtool', keyring, '--create-keyring',
           '--name=client.{}'.format(service), '--add-key={}'.format(key)]
    check_call(cmd)
    log('Created new ceph keyring at %s.' % keyring, level=DEBUG)


def create_keyring(service, key):
    """Deprecated. Please use the more accurately named 'add_key'"""
    return add_key(service, key)


def delete_keyring(service):
    """Delete an existing Ceph keyring."""
    keyring = _keyring_path(service)
    if not os.path.exists(keyring):
        log('Keyring does not exist at %s' % keyring, level=WARNING)
        return

    os.remove(keyring)
    log('Deleted ring at %s.' % keyring, level=INFO)


def create_key_file(service, key):
    """Create a file containing key."""
    keyfile = _keyfile_path(service)
    if os.path.exists(keyfile):
        log('Keyfile exists at %s.' % keyfile, level=WARNING)
        return

    with open(keyfile, 'w') as fd:
        fd.write(key)

    log('Created new keyfile at %s.' % keyfile, level=INFO)


def get_ceph_nodes(relation='ceph'):
    """Query named relation to determine current nodes."""
    hosts = []
    for r_id in relation_ids(relation):
        for unit in related_units(r_id):
            hosts.append(relation_get('private-address', unit=unit, rid=r_id))

    return hosts


def configure(service, key, auth, use_syslog):
    """Perform basic configuration of Ceph."""
    add_key(service, key)
    create_key_file(service, key)
    hosts = get_ceph_nodes()
    with open('/etc/ceph/ceph.conf', 'w') as ceph_conf:
        ceph_conf.write(CEPH_CONF.format(auth=auth,
                                         keyring=_keyring_path(service),
                                         mon_hosts=",".join(map(str, hosts)),
                                         use_syslog=use_syslog))
    modprobe('rbd')


def image_mapped(name):
    """Determine whether a RADOS block device is mapped locally."""
    try:
        out = check_output(['rbd', 'showmapped'])
        if six.PY3:
            out = out.decode('UTF-8')
    except CalledProcessError:
        return False

    return name in out


def map_block_storage(service, pool, image):
    """Map a RADOS block device for local use."""
    cmd = [
        'rbd',
        'map',
        '{}/{}'.format(pool, image),
        '--user',
        service,
        '--secret',
        _keyfile_path(service),
    ]
    check_call(cmd)


def filesystem_mounted(fs):
    """Determine whether a filesystem is already mounted."""
    return fs in [f for f, m in mounts()]


def make_filesystem(blk_device, fstype='ext4', timeout=10):
    """Make a new filesystem on the specified block device."""
    count = 0
    e_noent = errno.ENOENT
    while not os.path.exists(blk_device):
        if count >= timeout:
            log('Gave up waiting on block device %s' % blk_device,
                level=ERROR)
            raise IOError(e_noent, os.strerror(e_noent), blk_device)

        log('Waiting for block device %s to appear' % blk_device,
            level=DEBUG)
        count += 1
        time.sleep(1)
    else:
        log('Formatting block device %s as filesystem %s.' %
            (blk_device, fstype), level=INFO)
        check_call(['mkfs', '-t', fstype, blk_device])


def place_data_on_block_device(blk_device, data_src_dst):
    """Migrate data in data_src_dst to blk_device and then remount."""
    # mount block device into /mnt
    mount(blk_device, '/mnt')
    # copy data to /mnt
    copy_files(data_src_dst, '/mnt')
    # umount block device
    umount('/mnt')
    # Grab user/group ID's from original source
    _dir = os.stat(data_src_dst)
    uid = _dir.st_uid
    gid = _dir.st_gid
    # re-mount where the data should originally be
    # TODO: persist is currently a NO-OP in core.host
    mount(blk_device, data_src_dst, persist=True)
    # ensure original ownership of new mount.
    os.chown(data_src_dst, uid, gid)


def copy_files(src, dst, symlinks=False, ignore=None):
    """Copy files from src to dst."""
    for item in os.listdir(src):
        s = os.path.join(src, item)
        d = os.path.join(dst, item)
        if os.path.isdir(s):
            shutil.copytree(s, d, symlinks, ignore)
        else:
            shutil.copy2(s, d)


def ensure_ceph_storage(service, pool, rbd_img, sizemb, mount_point,
                        blk_device, fstype, system_services=[],
                        replicas=3):
    """NOTE: This function must only be called from a single service unit for
    the same rbd_img otherwise data loss will occur.

    Ensures given pool and RBD image exists, is mapped to a block device,
    and the device is formatted and mounted at the given mount_point.

    If formatting a device for the first time, data existing at mount_point
    will be migrated to the RBD device before being re-mounted.

    All services listed in system_services will be stopped prior to data
    migration and restarted when complete.
    """
    # Ensure pool, RBD image, RBD mappings are in place.
    if not pool_exists(service, pool):
        log('Creating new pool {}.'.format(pool), level=INFO)
        create_pool(service, pool, replicas=replicas)

    if not rbd_exists(service, pool, rbd_img):
        log('Creating RBD image ({}).'.format(rbd_img), level=INFO)
        create_rbd_image(service, pool, rbd_img, sizemb)

    if not image_mapped(rbd_img):
        log('Mapping RBD Image {} as a Block Device.'.format(rbd_img),
            level=INFO)
        map_block_storage(service, pool, rbd_img)

    # make file system
    # TODO: What happens if for whatever reason this is run again and
    # the data is already in the rbd device and/or is mounted??
    # When it is mounted already, it will fail to make the fs
    # XXX: This is really sketchy!  Need to at least add an fstab entry
    #      otherwise this hook will blow away existing data if its executed
    #      after a reboot.
    if not filesystem_mounted(mount_point):
        make_filesystem(blk_device, fstype)

        for svc in system_services:
            if service_running(svc):
                log('Stopping services {} prior to migrating data.'
                    .format(svc), level=DEBUG)
                service_stop(svc)

        place_data_on_block_device(blk_device, mount_point)

        for svc in system_services:
            log('Starting service {} after migrating data.'
                .format(svc), level=DEBUG)
            service_start(svc)


def ensure_ceph_keyring(service, user=None, group=None,
                        relation='ceph', key=None):
    """Ensures a ceph keyring is created for a named service and optionally
    ensures user and group ownership.

    @returns boolean: Flag to indicate whether a key was successfully written
                      to disk based on either relation data or a supplied key
    """
    if not key:
        for rid in relation_ids(relation):
            for unit in related_units(rid):
                key = relation_get('key', rid=rid, unit=unit)
                if key:
                    break

    if not key:
        return False

    add_key(service=service, key=key)
    keyring = _keyring_path(service)
    if user and group:
        check_call(['chown', '%s.%s' % (user, group), keyring])

    return True


class CephBrokerRq(object):
    """Ceph broker request.

    Multiple operations can be added to a request and sent to the Ceph broker
    to be executed.

    Request is json-encoded for sending over the wire.

    The API is versioned and defaults to version 1.
    """

    def __init__(self, api_version=1, request_id=None, raw_request_data=None):
        """Initialize CephBrokerRq object.

        Builds a new empty request or rebuilds a request from on-wire JSON
        data.

        :param api_version: API version for request (default: 1).
        :type api_version: Optional[int]
        :param request_id: Unique identifier for request.
                           (default: string representation of generated UUID)
        :type request_id: Optional[str]
        :param raw_request_data: JSON-encoded string to build request from.
        :type raw_request_data: Optional[str]
        :raises: KeyError
        """
        if raw_request_data:
            request_data = json.loads(raw_request_data)
            self.api_version = request_data['api-version']
            self.request_id = request_data['request-id']
            self.set_ops(request_data['ops'])
        else:
            self.api_version = api_version
            if request_id:
                self.request_id = request_id
            else:
                self.request_id = str(uuid.uuid1())
            self.ops = []

    def add_op(self, op):
        """Add an op if it is not already in the list.

        :param op: Operation to add.
        :type op: dict
        """
        if op not in self.ops:
            self.ops.append(op)

    def add_op_request_access_to_group(self, name, namespace=None,
                                       permission=None, key_name=None,
                                       object_prefix_permissions=None):
        """
        Adds the requested permissions to the current service's Ceph key,
        allowing the key to access only the specified pools or
        object prefixes. object_prefix_permissions should be a dictionary
        keyed on the permission with the corresponding value being a list
        of prefixes to apply that permission to.
            {
                'rwx': ['prefix1', 'prefix2'],
                'class-read': ['prefix3']}
        """
        self.add_op({
            'op': 'add-permissions-to-key', 'group': name,
            'namespace': namespace,
            'name': key_name or service_name(),
            'group-permission': permission,
            'object-prefix-permissions': object_prefix_permissions})

    def add_op_create_pool(self, name, replica_count=3, pg_num=None,
                           weight=None, group=None, namespace=None,
                           app_name=None, max_bytes=None, max_objects=None):
        """DEPRECATED: Use ``add_op_create_replicated_pool()`` or
                       ``add_op_create_erasure_pool()`` instead.
        """
        return self.add_op_create_replicated_pool(
            name, replica_count=replica_count, pg_num=pg_num, weight=weight,
            group=group, namespace=namespace, app_name=app_name,
            max_bytes=max_bytes, max_objects=max_objects)

    # Use function parameters and docstring to define types in a compatible
    # manner.
    #
    # NOTE: Our caller should always use a kwarg Dict when calling us so
    # no need to maintain fixed order/position for parameters. Please keep them
    # sorted by name when adding new ones.
    def _partial_build_common_op_create(self,
                                        app_name=None,
                                        compression_algorithm=None,
                                        compression_mode=None,
                                        compression_required_ratio=None,
                                        compression_min_blob_size=None,
                                        compression_min_blob_size_hdd=None,
                                        compression_min_blob_size_ssd=None,
                                        compression_max_blob_size=None,
                                        compression_max_blob_size_hdd=None,
                                        compression_max_blob_size_ssd=None,
                                        group=None,
                                        max_bytes=None,
                                        max_objects=None,
                                        namespace=None,
                                        rbd_mirroring_mode='pool',
                                        weight=None):
        """Build common part of a create pool operation.

        :param app_name: Tag pool with application name. Note that there is
                         certain protocols emerging upstream with regard to
                         meaningful application names to use.
                         Examples are 'rbd' and 'rgw'.
        :type app_name: Optional[str]
        :param compression_algorithm: Compressor to use, one of:
                                      ('lz4', 'snappy', 'zlib', 'zstd')
        :type compression_algorithm: Optional[str]
        :param compression_mode: When to compress data, one of:
                                 ('none', 'passive', 'aggressive', 'force')
        :type compression_mode: Optional[str]
        :param compression_required_ratio: Minimum compression ratio for data
                                           chunk, if the requested ratio is not
                                           achieved the compressed version will
                                           be thrown away and the original
                                           stored.
        :type compression_required_ratio: Optional[float]
        :param compression_min_blob_size: Chunks smaller than this are never
                                          compressed (unit: bytes).
        :type compression_min_blob_size: Optional[int]
        :param compression_min_blob_size_hdd: Chunks smaller than this are not
                                              compressed when destined to
                                              rotational media (unit: bytes).
        :type compression_min_blob_size_hdd: Optional[int]
        :param compression_min_blob_size_ssd: Chunks smaller than this are not
                                              compressed when destined to flash
                                              media (unit: bytes).
        :type compression_min_blob_size_ssd: Optional[int]
        :param compression_max_blob_size: Chunks larger than this are broken
                                          into N * compression_max_blob_size
                                          chunks before being compressed
                                          (unit: bytes).
        :type compression_max_blob_size: Optional[int]
        :param compression_max_blob_size_hdd: Chunks larger than this are
                                              broken into
                                              N * compression_max_blob_size_hdd
                                              chunks before being compressed
                                              when destined for rotational
                                              media (unit: bytes)
        :type compression_max_blob_size_hdd: Optional[int]
        :param compression_max_blob_size_ssd: Chunks larger than this are
                                              broken into
                                              N * compression_max_blob_size_ssd
                                              chunks before being compressed
                                              when destined for flash media
                                              (unit: bytes).
        :type compression_max_blob_size_ssd: Optional[int]
        :param group: Group to add pool to
        :type group: Optional[str]
        :param max_bytes: Maximum bytes quota to apply
        :type max_bytes: Optional[int]
        :param max_objects: Maximum objects quota to apply
        :type max_objects: Optional[int]
        :param namespace: Group namespace
        :type namespace: Optional[str]
        :param rbd_mirroring_mode: Pool mirroring mode used when Ceph RBD
                                   mirroring is enabled.
        :type rbd_mirroring_mode: Optional[str]
        :param weight: The percentage of data that is expected to be contained
                       in the pool from the total available space on the OSDs.
                       Used to calculate number of Placement Groups to create
                       for pool.
        :type weight: Optional[float]
        :returns: Dictionary with kwarg name as key.
        :rtype: Dict[str,any]
        :raises: AssertionError
        """
        return {
            'app-name': app_name,
            'compression-algorithm': compression_algorithm,
            'compression-mode': compression_mode,
            'compression-required-ratio': compression_required_ratio,
            'compression-min-blob-size': compression_min_blob_size,
            'compression-min-blob-size-hdd': compression_min_blob_size_hdd,
            'compression-min-blob-size-ssd': compression_min_blob_size_ssd,
            'compression-max-blob-size': compression_max_blob_size,
            'compression-max-blob-size-hdd': compression_max_blob_size_hdd,
            'compression-max-blob-size-ssd': compression_max_blob_size_ssd,
            'group': group,
            'max-bytes': max_bytes,
            'max-objects': max_objects,
            'group-namespace': namespace,
            'rbd-mirroring-mode': rbd_mirroring_mode,
            'weight': weight,
        }

    def add_op_create_replicated_pool(self, name, replica_count=3, pg_num=None,
                                      **kwargs):
        """Adds an operation to create a replicated pool.

        Refer to docstring for ``_partial_build_common_op_create`` for
        documentation of keyword arguments.

        :param name: Name of pool to create
        :type name: str
        :param replica_count: Number of copies Ceph should keep of your data.
        :type replica_count: int
        :param pg_num: Request specific number of Placement Groups to create
                       for pool.
        :type pg_num: int
        :raises: AssertionError if provided data is of invalid type/range
        """
        if pg_num and kwargs.get('weight'):
            raise ValueError('pg_num and weight are mutually exclusive')

        op = {
            'op': 'create-pool',
            'name': name,
            'replicas': replica_count,
            'pg_num': pg_num,
        }
        op.update(self._partial_build_common_op_create(**kwargs))

        # Initialize Pool-object to validate type and range of ops.
        pool = ReplicatedPool('dummy-service', op=op)
        pool.validate()

        self.add_op(op)

    def add_op_create_erasure_pool(self, name, erasure_profile=None,
                                   allow_ec_overwrites=False, **kwargs):
        """Adds an operation to create a erasure coded pool.

        Refer to docstring for ``_partial_build_common_op_create`` for
        documentation of keyword arguments.

        :param name: Name of pool to create
        :type name: str
        :param erasure_profile: Name of erasure code profile to use.  If not
                                set the ceph-mon unit handling the broker
                                request will set its default value.
        :type erasure_profile: str
        :param allow_ec_overwrites: allow EC pools to be overridden
        :type allow_ec_overwrites: bool
        :raises: AssertionError if provided data is of invalid type/range
        """
        op = {
            'op': 'create-pool',
            'name': name,
            'pool-type': 'erasure',
            'erasure-profile': erasure_profile,
            'allow-ec-overwrites': allow_ec_overwrites,
        }
        op.update(self._partial_build_common_op_create(**kwargs))

        # Initialize Pool-object to validate type and range of ops.
        pool = ErasurePool('dummy-service', op)
        pool.validate()

        self.add_op(op)

    def add_op_create_erasure_profile(self, name,
                                      erasure_type='jerasure',
                                      erasure_technique=None,
                                      k=None, m=None,
                                      failure_domain=None,
                                      lrc_locality=None,
                                      shec_durability_estimator=None,
                                      clay_helper_chunks=None,
                                      device_class=None,
                                      clay_scalar_mds=None,
                                      lrc_crush_locality=None):
        """Adds an operation to create a erasure coding profile.

        :param name: Name of profile to create
        :type name: str
        :param erasure_type: Which of the erasure coding plugins should be used
        :type erasure_type: string
        :param erasure_technique: EC plugin technique to use
        :type erasure_technique: string
        :param k: Number of data chunks
        :type k: int
        :param m: Number of coding chunks
        :type m: int
        :param lrc_locality: Group the coding and data chunks into sets of size locality
                             (lrc plugin)
        :type lrc_locality: int
        :param durability_estimator: The number of parity chunks each of which includes
                                     a data chunk in its calculation range (shec plugin)
        :type durability_estimator: int
        :param helper_chunks: The number of helper chunks to use for recovery operations
                              (clay plugin)
        :type: helper_chunks: int
        :param failure_domain: Type of failure domain from Ceph bucket types
                               to be used
        :type failure_domain: string
        :param device_class: Device class to use for profile (ssd, hdd)
        :type device_class: string
        :param clay_scalar_mds: Plugin to use for CLAY layered construction
                                (jerasure|isa|shec)
        :type clay_scaler_mds: string
        :param lrc_crush_locality: Type of crush bucket in which set of chunks
                                   defined by lrc_locality will be stored.
        :type lrc_crush_locality: string
        """
        self.add_op({'op': 'create-erasure-profile',
                     'name': name,
                     'k': k,
                     'm': m,
                     'l': lrc_locality,
                     'c': shec_durability_estimator,
                     'd': clay_helper_chunks,
                     'erasure-type': erasure_type,
                     'erasure-technique': erasure_technique,
                     'failure-domain': failure_domain,
                     'device-class': device_class,
                     'scalar-mds': clay_scalar_mds,
                     'crush-locality': lrc_crush_locality})

    def set_ops(self, ops):
        """Set request ops to provided value.

        Useful for injecting ops that come from a previous request
        to allow comparisons to ensure validity.
        """
        self.ops = ops

    @property
    def request(self):
        return json.dumps({'api-version': self.api_version, 'ops': self.ops,
                           'request-id': self.request_id})

    def _ops_equal(self, other):
        keys_to_compare = [
            'replicas', 'name', 'op', 'pg_num', 'group-permission',
            'object-prefix-permissions',
        ]
        keys_to_compare += list(self._partial_build_common_op_create().keys())
        if len(self.ops) == len(other.ops):
            for req_no in range(0, len(self.ops)):
                for key in keys_to_compare:
                    if self.ops[req_no].get(key) != other.ops[req_no].get(key):
                        return False
        else:
            return False
        return True

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        if self.api_version == other.api_version and \
                self._ops_equal(other):
            return True
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)


class CephBrokerRsp(object):
    """Ceph broker response.

    Response is json-decoded and contents provided as methods/properties.

    The API is versioned and defaults to version 1.
    """

    def __init__(self, encoded_rsp):
        self.api_version = None
        self.rsp = json.loads(encoded_rsp)

    @property
    def request_id(self):
        return self.rsp.get('request-id')

    @property
    def exit_code(self):
        return self.rsp.get('exit-code')

    @property
    def exit_msg(self):
        return self.rsp.get('stderr')


# Ceph Broker Conversation:
# If a charm needs an action to be taken by ceph it can create a CephBrokerRq
# and send that request to ceph via the ceph relation. The CephBrokerRq has a
# unique id so that the client can identity which CephBrokerRsp is associated
# with the request. Ceph will also respond to each client unit individually
# creating a response key per client unit eg glance/0 will get a CephBrokerRsp
# via key broker-rsp-glance-0
#
# To use this the charm can just do something like:
#
# from charmhelpers.contrib.storage.linux.ceph import (
#     send_request_if_needed,
#     is_request_complete,
#     CephBrokerRq,
# )
#
# @hooks.hook('ceph-relation-changed')
# def ceph_changed():
#     rq = CephBrokerRq()
#     rq.add_op_create_pool(name='poolname', replica_count=3)
#
#     if is_request_complete(rq):
#         <Request complete actions>
#     else:
#         send_request_if_needed(get_ceph_request())
#
# CephBrokerRq and CephBrokerRsp are serialized into JSON. Below is an example
# of glance having sent a request to ceph which ceph has successfully processed
#  'ceph:8': {
#      'ceph/0': {
#          'auth': 'cephx',
#          'broker-rsp-glance-0': '{"request-id": "0bc7dc54", "exit-code": 0}',
#          'broker_rsp': '{"request-id": "0da543b8", "exit-code": 0}',
#          'ceph-public-address': '10.5.44.103',
#          'key': 'AQCLDttVuHXINhAAvI144CB09dYchhHyTUY9BQ==',
#          'private-address': '10.5.44.103',
#      },
#      'glance/0': {
#          'broker_req': ('{"api-version": 1, "request-id": "0bc7dc54", '
#                         '"ops": [{"replicas": 3, "name": "glance", '
#                         '"op": "create-pool"}]}'),
#          'private-address': '10.5.44.109',
#      },
#  }

def get_previous_request(rid):
    """Return the last ceph broker request sent on a given relation

    :param rid: Relation id to query for request
    :type rid: str
    :returns: CephBrokerRq object or None if relation data not found.
    :rtype: Optional[CephBrokerRq]
    """
    broker_req = relation_get(attribute='broker_req', rid=rid,
                              unit=local_unit())
    if broker_req:
        return CephBrokerRq(raw_request_data=broker_req)


def get_request_states(request, relation='ceph'):
    """Return a dict of requests per relation id with their corresponding
       completion state.

    This allows a charm, which has a request for ceph, to see whether there is
    an equivalent request already being processed and if so what state that
    request is in.

    @param request: A CephBrokerRq object
    """
    complete = []
    requests = {}
    for rid in relation_ids(relation):
        complete = False
        previous_request = get_previous_request(rid)
        if request == previous_request:
            sent = True
            complete = is_request_complete_for_rid(previous_request, rid)
        else:
            sent = False
            complete = False

        requests[rid] = {
            'sent': sent,
            'complete': complete,
        }

    return requests


def is_request_sent(request, relation='ceph'):
    """Check to see if a functionally equivalent request has already been sent

    Returns True if a similair request has been sent

    @param request: A CephBrokerRq object
    """
    states = get_request_states(request, relation=relation)
    for rid in states.keys():
        if not states[rid]['sent']:
            return False

    return True


def is_request_complete(request, relation='ceph'):
    """Check to see if a functionally equivalent request has already been
    completed

    Returns True if a similair request has been completed

    @param request: A CephBrokerRq object
    """
    states = get_request_states(request, relation=relation)
    for rid in states.keys():
        if not states[rid]['complete']:
            return False

    return True


def is_request_complete_for_rid(request, rid):
    """Check if a given request has been completed on the given relation

    @param request: A CephBrokerRq object
    @param rid: Relation ID
    """
    broker_key = get_broker_rsp_key()
    for unit in related_units(rid):
        rdata = relation_get(rid=rid, unit=unit)
        if rdata.get(broker_key):
            rsp = CephBrokerRsp(rdata.get(broker_key))
            if rsp.request_id == request.request_id:
                if not rsp.exit_code:
                    return True
        else:
            # The remote unit sent no reply targeted at this unit so either the
            # remote ceph cluster does not support unit targeted replies or it
            # has not processed our request yet.
            if rdata.get('broker_rsp'):
                request_data = json.loads(rdata['broker_rsp'])
                if request_data.get('request-id'):
                    log('Ignoring legacy broker_rsp without unit key as remote '
                        'service supports unit specific replies', level=DEBUG)
                else:
                    log('Using legacy broker_rsp as remote service does not '
                        'supports unit specific replies', level=DEBUG)
                    rsp = CephBrokerRsp(rdata['broker_rsp'])
                    if not rsp.exit_code:
                        return True

    return False


def get_broker_rsp_key():
    """Return broker response key for this unit

    This is the key that ceph is going to use to pass request status
    information back to this unit
    """
    return 'broker-rsp-' + local_unit().replace('/', '-')


def send_request_if_needed(request, relation='ceph'):
    """Send broker request if an equivalent request has not already been sent

    @param request: A CephBrokerRq object
    """
    if is_request_sent(request, relation=relation):
        log('Request already sent but not complete, not sending new request',
            level=DEBUG)
    else:
        for rid in relation_ids(relation):
            log('Sending request {}'.format(request.request_id), level=DEBUG)
            relation_set(relation_id=rid, broker_req=request.request)
            relation_set(relation_id=rid, relation_settings={'unit-name': local_unit()})


def has_broker_rsp(rid=None, unit=None):
    """Return True if the broker_rsp key is 'truthy' (i.e. set to something) in the relation data.

    :param rid: The relation to check (default of None means current relation)
    :type rid: Union[str, None]
    :param unit: The remote unit to check (default of None means current unit)
    :type unit: Union[str, None]
    :returns: True if broker key exists and is set to something 'truthy'
    :rtype: bool
    """
    rdata = relation_get(rid=rid, unit=unit) or {}
    broker_rsp = rdata.get(get_broker_rsp_key())
    return True if broker_rsp else False


def is_broker_action_done(action, rid=None, unit=None):
    """Check whether broker action has completed yet.

    @param action: name of action to be performed
    @returns True if action complete otherwise False
    """
    rdata = relation_get(rid=rid, unit=unit) or {}
    broker_rsp = rdata.get(get_broker_rsp_key())
    if not broker_rsp:
        return False

    rsp = CephBrokerRsp(broker_rsp)
    unit_name = local_unit().partition('/')[2]
    key = "unit_{}_ceph_broker_action.{}".format(unit_name, action)
    kvstore = kv()
    val = kvstore.get(key=key)
    if val and val == rsp.request_id:
        return True

    return False


def mark_broker_action_done(action, rid=None, unit=None):
    """Mark action as having been completed.

    @param action: name of action to be performed
    @returns None
    """
    rdata = relation_get(rid=rid, unit=unit) or {}
    broker_rsp = rdata.get(get_broker_rsp_key())
    if not broker_rsp:
        return

    rsp = CephBrokerRsp(broker_rsp)
    unit_name = local_unit().partition('/')[2]
    key = "unit_{}_ceph_broker_action.{}".format(unit_name, action)
    kvstore = kv()
    kvstore.set(key=key, value=rsp.request_id)
    kvstore.flush()


class CephConfContext(object):
    """Ceph config (ceph.conf) context.

    Supports user-provided Ceph configuration settings. Use can provide a
    dictionary as the value for the config-flags charm option containing
    Ceph configuration settings keyede by their section in ceph.conf.
    """
    def __init__(self, permitted_sections=None):
        self.permitted_sections = permitted_sections or []

    def __call__(self):
        conf = config('config-flags')
        if not conf:
            return {}

        conf = config_flags_parser(conf)
        if not isinstance(conf, dict):
            log("Provided config-flags is not a dictionary - ignoring",
                level=WARNING)
            return {}

        permitted = self.permitted_sections
        if permitted:
            diff = set(conf.keys()).difference(set(permitted))
            if diff:
                log("Config-flags contains invalid keys '%s' - they will be "
                    "ignored" % (', '.join(diff)), level=WARNING)

        ceph_conf = {}
        for key in conf:
            if permitted and key not in permitted:
                log("Ignoring key '%s'" % key, level=WARNING)
                continue

            ceph_conf[key] = conf[key]
        return ceph_conf


class CephOSDConfContext(CephConfContext):
    """Ceph config (ceph.conf) context.

    Consolidates settings from config-flags via CephConfContext with
    settings provided by the mons. The config-flag values are preserved in
    conf['osd'], settings from the mons which do not clash with config-flag
    settings are in conf['osd_from_client'] and finally settings which do
    clash are in conf['osd_from_client_conflict']. Rather than silently drop
    the conflicting settings they are provided in the context so they can be
    rendered commented out to give some visibility to the admin.
    """

    def __init__(self, permitted_sections=None):
        super(CephOSDConfContext, self).__init__(
            permitted_sections=permitted_sections)
        try:
            self.settings_from_mons = get_osd_settings('mon')
        except OSDSettingConflict:
            log(
                "OSD settings from mons are inconsistent, ignoring them",
                level=WARNING)
            self.settings_from_mons = {}

    def filter_osd_from_mon_settings(self):
        """Filter settings from client relation against config-flags.

        :returns: A tuple (
            ,config-flag values,
            ,client settings which do not conflict with config-flag values,
            ,client settings which confilct with config-flag values)
        :rtype: (OrderedDict, OrderedDict, OrderedDict)
        """
        ceph_conf = super(CephOSDConfContext, self).__call__()
        conflicting_entries = {}
        clear_entries = {}
        for key, value in self.settings_from_mons.items():
            if key in ceph_conf.get('osd', {}):
                if ceph_conf['osd'][key] != value:
                    conflicting_entries[key] = value
            else:
                clear_entries[key] = value
        clear_entries = _order_dict_by_key(clear_entries)
        conflicting_entries = _order_dict_by_key(conflicting_entries)
        return ceph_conf, clear_entries, conflicting_entries

    def __call__(self):
        """Construct OSD config context.

        Standard context with two additional special keys.
            osd_from_client_conflict: client settings which confilct with
                                      config-flag values
            osd_from_client: settings which do not conflict with config-flag
                             values

        :returns: OSD config context dict.
        :rtype: dict
        """
        conf, osd_clear, osd_conflict = self.filter_osd_from_mon_settings()
        conf['osd_from_client_conflict'] = osd_conflict
        conf['osd_from_client'] = osd_clear
        return conf



# Copyright 2016 Canonical Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from charmhelpers.contrib.hardening.audits.file import (
    DirectoryPermissionAudit,
    TemplatedFile,
)
from charmhelpers.contrib.hardening.host import TEMPLATES_DIR
from charmhelpers.contrib.hardening import utils


def get_audits():
    """Get OS hardening security limits audits.

    :returns:  dictionary of audits
    """
    audits = []
    settings = utils.get_settings('os')

    # Ensure that the /etc/security/limits.d directory is only writable
    # by the root user, but others can execute and read.
    audits.append(DirectoryPermissionAudit('/etc/security/limits.d',
                                           user='root', group='root',
                                           mode=0o755))

    # If core dumps are not enabled, then don't allow core dumps to be
    # created as they may contain sensitive information.
    if not settings['security']['kernel_enable_core_dump']:
        audits.append(TemplatedFile('/etc/security/limits.d/10.hardcore.conf',
                                    SecurityLimitsContext(),
                                    template_dir=TEMPLATES_DIR,
                                    user='root', group='root', mode=0o0440))
    return audits


class SecurityLimitsContext(object):

    def __call__(self):
        settings = utils.get_settings('os')
        ctxt = {'disable_core_dump':
                not settings['security']['kernel_enable_core_dump']}
        return ctxt



# Copyright 2014-2021 Canonical Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#
# Copyright 2012 Canonical Ltd.
#
# Authors:
#  James Page <james.page@ubuntu.com>
#  Adam Gandelman <adamg@ubuntu.com>
#

"""
Helpers for clustering and determining "cluster leadership" and other
clustering-related helpers.
"""

import functools
import subprocess
import os
import time

from socket import gethostname as get_unit_hostname

import six

from charmhelpers.core.hookenv import (
    log,
    relation_ids,
    related_units as relation_list,
    relation_get,
    config as config_get,
    INFO,
    DEBUG,
    WARNING,
    unit_get,
    is_leader as juju_is_leader,
    status_set,
)
from charmhelpers.core.host import (
    modulo_distribution,
)
from charmhelpers.core.decorators import (
    retry_on_exception,
)
from charmhelpers.core.strutils import (
    bool_from_string,
)

DC_RESOURCE_NAME = 'DC'


class HAIncompleteConfig(Exception):
    pass


class HAIncorrectConfig(Exception):
    pass


class CRMResourceNotFound(Exception):
    pass


class CRMDCNotFound(Exception):
    pass


def is_elected_leader(resource):
    """
    Returns True if the charm executing this is the elected cluster leader.

    It relies on two mechanisms to determine leadership:
        1. If juju is sufficiently new and leadership election is supported,
        the is_leader command will be used.
        2. If the charm is part of a corosync cluster, call corosync to
        determine leadership.
        3. If the charm is not part of a corosync cluster, the leader is
        determined as being "the alive unit with the lowest unit number". In
        other words, the oldest surviving unit.
    """
    try:
        return juju_is_leader()
    except NotImplementedError:
        log('Juju leadership election feature not enabled'
            ', using fallback support',
            level=WARNING)

    if is_clustered():
        if not is_crm_leader(resource):
            log('Deferring action to CRM leader.', level=INFO)
            return False
    else:
        peers = peer_units()
        if peers and not oldest_peer(peers):
            log('Deferring action to oldest service unit.', level=INFO)
            return False
    return True


def is_clustered():
    for r_id in (relation_ids('ha') or []):
        for unit in (relation_list(r_id) or []):
            clustered = relation_get('clustered',
                                     rid=r_id,
                                     unit=unit)
            if clustered:
                return True
    return False


def is_crm_dc():
    """
    Determine leadership by querying the pacemaker Designated Controller
    """
    cmd = ['crm', 'status']
    try:
        status = subprocess.check_output(cmd, stderr=subprocess.STDOUT)
        if not isinstance(status, six.text_type):
            status = six.text_type(status, "utf-8")
    except subprocess.CalledProcessError as ex:
        raise CRMDCNotFound(str(ex))

    current_dc = ''
    for line in status.split('\n'):
        if line.startswith('Current DC'):
            # Current DC: juju-lytrusty-machine-2 (168108163) - partition with quorum
            current_dc = line.split(':')[1].split()[0]
    if current_dc == get_unit_hostname():
        return True
    elif current_dc == 'NONE':
        raise CRMDCNotFound('Current DC: NONE')

    return False


@retry_on_exception(5, base_delay=2,
                    exc_type=(CRMResourceNotFound, CRMDCNotFound))
def is_crm_leader(resource, retry=False):
    """
    Returns True if the charm calling this is the elected corosync leader,
    as returned by calling the external "crm" command.

    We allow this operation to be retried to avoid the possibility of getting a
    false negative. See LP #1396246 for more info.
    """
    if resource == DC_RESOURCE_NAME:
        return is_crm_dc()
    cmd = ['crm', 'resource', 'show', resource]
    try:
        status = subprocess.check_output(cmd, stderr=subprocess.STDOUT)
        if not isinstance(status, six.text_type):
            status = six.text_type(status, "utf-8")
    except subprocess.CalledProcessError:
        status = None

    if status and get_unit_hostname() in status:
        return True

    if status and "resource %s is NOT running" % (resource) in status:
        raise CRMResourceNotFound("CRM resource %s not found" % (resource))

    return False


def is_leader(resource):
    log("is_leader is deprecated. Please consider using is_crm_leader "
        "instead.", level=WARNING)
    return is_crm_leader(resource)


def peer_units(peer_relation="cluster"):
    peers = []
    for r_id in (relation_ids(peer_relation) or []):
        for unit in (relation_list(r_id) or []):
            peers.append(unit)
    return peers


def peer_ips(peer_relation='cluster', addr_key='private-address'):
    '''Return a dict of peers and their private-address'''
    peers = {}
    for r_id in relation_ids(peer_relation):
        for unit in relation_list(r_id):
            peers[unit] = relation_get(addr_key, rid=r_id, unit=unit)
    return peers


def oldest_peer(peers):
    """Determines who the oldest peer is by comparing unit numbers."""
    local_unit_no = int(os.getenv('JUJU_UNIT_NAME').split('/')[1])
    for peer in peers:
        remote_unit_no = int(peer.split('/')[1])
        if remote_unit_no < local_unit_no:
            return False
    return True


def eligible_leader(resource):
    log("eligible_leader is deprecated. Please consider using "
        "is_elected_leader instead.", level=WARNING)
    return is_elected_leader(resource)


def https():
    '''
    Determines whether enough data has been provided in configuration
    or relation data to configure HTTPS
    .
    returns: boolean
    '''
    use_https = config_get('use-https')
    if use_https and bool_from_string(use_https):
        return True
    if config_get('ssl_cert') and config_get('ssl_key'):
        return True
    for r_id in relation_ids('certificates'):
        for unit in relation_list(r_id):
            ca = relation_get('ca', rid=r_id, unit=unit)
            if ca:
                return True
    for r_id in relation_ids('identity-service'):
        for unit in relation_list(r_id):
            # TODO - needs fixing for new helper as ssl_cert/key suffixes with CN
            rel_state = [
                relation_get('https_keystone', rid=r_id, unit=unit),
                relation_get('ca_cert', rid=r_id, unit=unit),
            ]
            # NOTE: works around (LP: #1203241)
            if (None not in rel_state) and ('' not in rel_state):
                return True
    return False


def determine_api_port(public_port, singlenode_mode=False):
    '''
    Determine correct API server listening port based on
    existence of HTTPS reverse proxy and/or haproxy.

    public_port: int: standard public port for given service

    singlenode_mode: boolean: Shuffle ports when only a single unit is present

    returns: int: the correct listening port for the API service
    '''
    i = 0
    if singlenode_mode:
        i += 1
    elif len(peer_units()) > 0 or is_clustered():
        i += 1
    if https():
        i += 1
    return public_port - (i * 10)


def determine_apache_port(public_port, singlenode_mode=False):
    '''
    Description: Determine correct apache listening port based on public IP +
    state of the cluster.

    public_port: int: standard public port for given service

    singlenode_mode: boolean: Shuffle ports when only a single unit is present

    returns: int: the correct listening port for the HAProxy service
    '''
    i = 0
    if singlenode_mode:
        i += 1
    elif len(peer_units()) > 0 or is_clustered():
        i += 1
    return public_port - (i * 10)


determine_apache_port_single = functools.partial(
    determine_apache_port, singlenode_mode=True)


def get_hacluster_config(exclude_keys=None):
    '''
    Obtains all relevant configuration from charm configuration required
    for initiating a relation to hacluster:

        ha-bindiface, ha-mcastport, vip, os-internal-hostname,
        os-admin-hostname, os-public-hostname, os-access-hostname

    param: exclude_keys: list of setting key(s) to be excluded.
    returns: dict: A dict containing settings keyed by setting name.
    raises: HAIncompleteConfig if settings are missing or incorrect.
    '''
    settings = ['ha-bindiface', 'ha-mcastport', 'vip', 'os-internal-hostname',
                'os-admin-hostname', 'os-public-hostname', 'os-access-hostname']
    conf = {}
    for setting in settings:
        if exclude_keys and setting in exclude_keys:
            continue

        conf[setting] = config_get(setting)

    if not valid_hacluster_config():
        raise HAIncorrectConfig('Insufficient or incorrect config data to '
                                'configure hacluster.')
    return conf


def valid_hacluster_config():
    '''
    Check that either vip or dns-ha is set. If dns-ha then one of os-*-hostname
    must be set.

    Note: ha-bindiface and ha-macastport both have defaults and will always
    be set. We only care that either vip or dns-ha is set.

    :returns: boolean: valid config returns true.
    raises: HAIncompatibileConfig if settings conflict.
    raises: HAIncompleteConfig if settings are missing.
    '''
    vip = config_get('vip')
    dns = config_get('dns-ha')
    if not(bool(vip) ^ bool(dns)):
        msg = ('HA: Either vip or dns-ha must be set but not both in order to '
               'use high availability')
        status_set('blocked', msg)
        raise HAIncorrectConfig(msg)

    # If dns-ha then one of os-*-hostname must be set
    if dns:
        dns_settings = ['os-internal-hostname', 'os-admin-hostname',
                        'os-public-hostname', 'os-access-hostname']
        # At this point it is unknown if one or all of the possible
        # network spaces are in HA. Validate at least one is set which is
        # the minimum required.
        for setting in dns_settings:
            if config_get(setting):
                log('DNS HA: At least one hostname is set {}: {}'
                    ''.format(setting, config_get(setting)),
                    level=DEBUG)
                return True

        msg = ('DNS HA: At least one os-*-hostname(s) must be set to use '
               'DNS HA')
        status_set('blocked', msg)
        raise HAIncompleteConfig(msg)

    log('VIP HA: VIP is set {}'.format(vip), level=DEBUG)
    return True


def canonical_url(configs, vip_setting='vip'):
    '''
    Returns the correct HTTP URL to this host given the state of HTTPS
    configuration and hacluster.

    :configs    : OSTemplateRenderer: A config tempating object to inspect for
                                      a complete https context.

    :vip_setting:                str: Setting in charm config that specifies
                                      VIP address.
    '''
    scheme = 'http'
    if 'https' in configs.complete_contexts():
        scheme = 'https'
    if is_clustered():
        addr = config_get(vip_setting)
    else:
        addr = unit_get('private-address')
    return '%s://%s' % (scheme, addr)


def distributed_wait(modulo=None, wait=None, operation_name='operation'):
    ''' Distribute operations by waiting based on modulo_distribution

    If modulo and or wait are not set, check config_get for those values.
    If config values are not set, default to modulo=3 and wait=30.

    :param modulo: int The modulo number creates the group distribution
    :param wait: int The constant time wait value
    :param operation_name: string Operation name for status message
                           i.e.  'restart'
    :side effect: Calls config_get()
    :side effect: Calls log()
    :side effect: Calls status_set()
    :side effect: Calls time.sleep()
    '''
    if modulo is None:
        modulo = config_get('modulo-nodes') or 3
    if wait is None:
        wait = config_get('known-wait') or 30
    if juju_is_leader():
        # The leader should never wait
        calculated_wait = 0
    else:
        # non_zero_wait=True guarantees the non-leader who gets modulo 0
        # will still wait
        calculated_wait = modulo_distribution(modulo=modulo, wait=wait,
                                              non_zero_wait=True)
    msg = "Waiting {} seconds for {} ...".format(calculated_wait,
                                                 operation_name)
    log(msg, DEBUG)
    status_set('maintenance', msg)
    time.sleep(calculated_wait)


def get_managed_services_and_ports(services, external_ports,
                                   external_services=None,
                                   port_conv_f=determine_apache_port_single):
    """Get the services and ports managed by this charm.

    Return only the services and corresponding ports that are managed by this
    charm. This excludes haproxy when there is a relation with hacluster. This
    is because this charm passes responsibility for stopping and starting
    haproxy to hacluster.

    Similarly, if a relation with hacluster exists then the ports returned by
    this method correspond to those managed by the apache server rather than
    haproxy.

    :param services: List of services.
    :type services: List[str]
    :param external_ports: List of ports managed by external services.
    :type external_ports: List[int]
    :param external_services: List of services to be removed if ha relation is
                              present.
    :type external_services: List[str]
    :param port_conv_f: Function to apply to ports to calculate the ports
                        managed by services controlled by this charm.
    :type port_convert_func: f()
    :returns: A tuple containing a list of services first followed by a list of
              ports.
    :rtype: Tuple[List[str], List[int]]
    """
    if external_services is None:
        external_services = ['haproxy']
    if relation_ids('ha'):
        for svc in external_services:
            try:
                services.remove(svc)
            except ValueError:
                pass
        external_ports = [port_conv_f(p) for p in external_ports]
    return services, external_ports



# Copyright 2014-2021 Canonical Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tools for working with the host system"""
# Copyright 2012 Canonical Ltd.
#
# Authors:
#  Nick Moffitt <nick.moffitt@canonical.com>
#  Matthew Wedgwood <matthew.wedgwood@canonical.com>

import errno
import os
import re
import pwd
import glob
import grp
import random
import string
import subprocess
import hashlib
import functools
import itertools
import six

from contextlib import contextmanager
from collections import OrderedDict, defaultdict
from .hookenv import log, INFO, DEBUG, local_unit, charm_name
from .fstab import Fstab
from charmhelpers.osplatform import get_platform

__platform__ = get_platform()
if __platform__ == "ubuntu":
    from charmhelpers.core.host_factory.ubuntu import (  # NOQA:F401
        service_available,
        add_new_group,
        lsb_release,
        cmp_pkgrevno,
        CompareHostReleases,
        get_distrib_codename,
        arch
    )  # flake8: noqa -- ignore F401 for this import
elif __platform__ == "centos":
    from charmhelpers.core.host_factory.centos import (  # NOQA:F401
        service_available,
        add_new_group,
        lsb_release,
        cmp_pkgrevno,
        CompareHostReleases,
    )  # flake8: noqa -- ignore F401 for this import

UPDATEDB_PATH = '/etc/updatedb.conf'
CA_CERT_DIR = '/usr/local/share/ca-certificates'


def service_start(service_name, **kwargs):
    """Start a system service.

    The specified service name is managed via the system level init system.
    Some init systems (e.g. upstart) require that additional arguments be
    provided in order to directly control service instances whereas other init
    systems allow for addressing instances of a service directly by name (e.g.
    systemd).

    The kwargs allow for the additional parameters to be passed to underlying
    init systems for those systems which require/allow for them. For example,
    the ceph-osd upstart script requires the id parameter to be passed along
    in order to identify which running daemon should be reloaded. The follow-
    ing example stops the ceph-osd service for instance id=4:

    service_stop('ceph-osd', id=4)

    :param service_name: the name of the service to stop
    :param **kwargs: additional parameters to pass to the init system when
                     managing services. These will be passed as key=value
                     parameters to the init system's commandline. kwargs
                     are ignored for systemd enabled systems.
    """
    return service('start', service_name, **kwargs)


def service_stop(service_name, **kwargs):
    """Stop a system service.

    The specified service name is managed via the system level init system.
    Some init systems (e.g. upstart) require that additional arguments be
    provided in order to directly control service instances whereas other init
    systems allow for addressing instances of a service directly by name (e.g.
    systemd).

    The kwargs allow for the additional parameters to be passed to underlying
    init systems for those systems which require/allow for them. For example,
    the ceph-osd upstart script requires the id parameter to be passed along
    in order to identify which running daemon should be reloaded. The follow-
    ing example stops the ceph-osd service for instance id=4:

    service_stop('ceph-osd', id=4)

    :param service_name: the name of the service to stop
    :param **kwargs: additional parameters to pass to the init system when
                     managing services. These will be passed as key=value
                     parameters to the init system's commandline. kwargs
                     are ignored for systemd enabled systems.
    """
    return service('stop', service_name, **kwargs)


def service_restart(service_name, **kwargs):
    """Restart a system service.

    The specified service name is managed via the system level init system.
    Some init systems (e.g. upstart) require that additional arguments be
    provided in order to directly control service instances whereas other init
    systems allow for addressing instances of a service directly by name (e.g.
    systemd).

    The kwargs allow for the additional parameters to be passed to underlying
    init systems for those systems which require/allow for them. For example,
    the ceph-osd upstart script requires the id parameter to be passed along
    in order to identify which running daemon should be restarted. The follow-
    ing example restarts the ceph-osd service for instance id=4:

    service_restart('ceph-osd', id=4)

    :param service_name: the name of the service to restart
    :param **kwargs: additional parameters to pass to the init system when
                     managing services. These will be passed as key=value
                     parameters to the  init system's commandline. kwargs
                     are ignored for init systems not allowing additional
                     parameters via the commandline (systemd).
    """
    return service('restart', service_name)


def service_reload(service_name, restart_on_failure=False, **kwargs):
    """Reload a system service, optionally falling back to restart if
    reload fails.

    The specified service name is managed via the system level init system.
    Some init systems (e.g. upstart) require that additional arguments be
    provided in order to directly control service instances whereas other init
    systems allow for addressing instances of a service directly by name (e.g.
    systemd).

    The kwargs allow for the additional parameters to be passed to underlying
    init systems for those systems which require/allow for them. For example,
    the ceph-osd upstart script requires the id parameter to be passed along
    in order to identify which running daemon should be reloaded. The follow-
    ing example restarts the ceph-osd service for instance id=4:

    service_reload('ceph-osd', id=4)

    :param service_name: the name of the service to reload
    :param restart_on_failure: boolean indicating whether to fallback to a
                               restart if the reload fails.
    :param **kwargs: additional parameters to pass to the init system when
                     managing services. These will be passed as key=value
                     parameters to the  init system's commandline. kwargs
                     are ignored for init systems not allowing additional
                     parameters via the commandline (systemd).
    """
    service_result = service('reload', service_name, **kwargs)
    if not service_result and restart_on_failure:
        service_result = service('restart', service_name, **kwargs)
    return service_result


def service_pause(service_name, init_dir="/etc/init", initd_dir="/etc/init.d",
                  **kwargs):
    """Pause a system service.

    Stop it, and prevent it from starting again at boot.

    :param service_name: the name of the service to pause
    :param init_dir: path to the upstart init directory
    :param initd_dir: path to the sysv init directory
    :param **kwargs: additional parameters to pass to the init system when
                     managing services. These will be passed as key=value
                     parameters to the init system's commandline. kwargs
                     are ignored for init systems which do not support
                     key=value arguments via the commandline.
    """
    stopped = True
    if service_running(service_name, **kwargs):
        stopped = service_stop(service_name, **kwargs)
    upstart_file = os.path.join(init_dir, "{}.conf".format(service_name))
    sysv_file = os.path.join(initd_dir, service_name)
    if init_is_systemd(service_name=service_name):
        service('disable', service_name)
        service('mask', service_name)
    elif os.path.exists(upstart_file):
        override_path = os.path.join(
            init_dir, '{}.override'.format(service_name))
        with open(override_path, 'w') as fh:
            fh.write("manual\n")
    elif os.path.exists(sysv_file):
        subprocess.check_call(["update-rc.d", service_name, "disable"])
    else:
        raise ValueError(
            "Unable to detect {0} as SystemD, Upstart {1} or"
            " SysV {2}".format(
                service_name, upstart_file, sysv_file))
    return stopped


def service_resume(service_name, init_dir="/etc/init",
                   initd_dir="/etc/init.d", **kwargs):
    """Resume a system service.

    Re-enable starting again at boot. Start the service.

    :param service_name: the name of the service to resume
    :param init_dir: the path to the init dir
    :param initd dir: the path to the initd dir
    :param **kwargs: additional parameters to pass to the init system when
                     managing services. These will be passed as key=value
                     parameters to the init system's commandline. kwargs
                     are ignored for systemd enabled systems.
    """
    upstart_file = os.path.join(init_dir, "{}.conf".format(service_name))
    sysv_file = os.path.join(initd_dir, service_name)
    if init_is_systemd(service_name=service_name):
        service('unmask', service_name)
        service('enable', service_name)
    elif os.path.exists(upstart_file):
        override_path = os.path.join(
            init_dir, '{}.override'.format(service_name))
        if os.path.exists(override_path):
            os.unlink(override_path)
    elif os.path.exists(sysv_file):
        subprocess.check_call(["update-rc.d", service_name, "enable"])
    else:
        raise ValueError(
            "Unable to detect {0} as SystemD, Upstart {1} or"
            " SysV {2}".format(
                service_name, upstart_file, sysv_file))
    started = service_running(service_name, **kwargs)

    if not started:
        started = service_start(service_name, **kwargs)
    return started


def service(action, service_name, **kwargs):
    """Control a system service.

    :param action: the action to take on the service
    :param service_name: the name of the service to perform th action on
    :param **kwargs: additional params to be passed to the service command in
                    the form of key=value.
    """
    if init_is_systemd(service_name=service_name):
        cmd = ['systemctl', action, service_name]
    else:
        cmd = ['service', service_name, action]
        for key, value in six.iteritems(kwargs):
            parameter = '%s=%s' % (key, value)
            cmd.append(parameter)
    return subprocess.call(cmd) == 0


_UPSTART_CONF = "/etc/init/{}.conf"
_INIT_D_CONF = "/etc/init.d/{}"


def service_running(service_name, **kwargs):
    """Determine whether a system service is running.

    :param service_name: the name of the service
    :param **kwargs: additional args to pass to the service command. This is
                     used to pass additional key=value arguments to the
                     service command line for managing specific instance
                     units (e.g. service ceph-osd status id=2). The kwargs
                     are ignored in systemd services.
    """
    if init_is_systemd(service_name=service_name):
        return service('is-active', service_name)
    else:
        if os.path.exists(_UPSTART_CONF.format(service_name)):
            try:
                cmd = ['status', service_name]
                for key, value in six.iteritems(kwargs):
                    parameter = '%s=%s' % (key, value)
                    cmd.append(parameter)
                output = subprocess.check_output(
                    cmd, stderr=subprocess.STDOUT).decode('UTF-8')
            except subprocess.CalledProcessError:
                return False
            else:
                # This works for upstart scripts where the 'service' command
                # returns a consistent string to represent running
                # 'start/running'
                if ("start/running" in output or
                        "is running" in output or
                        "up and running" in output):
                    return True
        elif os.path.exists(_INIT_D_CONF.format(service_name)):
            # Check System V scripts init script return codes
            return service('status', service_name)
        return False


SYSTEMD_SYSTEM = '/run/systemd/system'


def init_is_systemd(service_name=None):
    """
    Returns whether the host uses systemd for the specified service.

    @param Optional[str] service_name: specific name of service
    """
    if str(service_name).startswith("snap."):
        return True
    if lsb_release()['DISTRIB_CODENAME'] == 'trusty':
        return False
    return os.path.isdir(SYSTEMD_SYSTEM)


def adduser(username, password=None, shell='/bin/bash',
            system_user=False, primary_group=None,
            secondary_groups=None, uid=None, home_dir=None):
    """Add a user to the system.

    Will log but otherwise succeed if the user already exists.

    :param str username: Username to create
    :param str password: Password for user; if ``None``, create a system user
    :param str shell: The default shell for the user
    :param bool system_user: Whether to create a login or system user
    :param str primary_group: Primary group for user; defaults to username
    :param list secondary_groups: Optional list of additional groups
    :param int uid: UID for user being created
    :param str home_dir: Home directory for user

    :returns: The password database entry struct, as returned by `pwd.getpwnam`
    """
    try:
        user_info = pwd.getpwnam(username)
        log('user {0} already exists!'.format(username))
        if uid:
            user_info = pwd.getpwuid(int(uid))
            log('user with uid {0} already exists!'.format(uid))
    except KeyError:
        log('creating user {0}'.format(username))
        cmd = ['useradd']
        if uid:
            cmd.extend(['--uid', str(uid)])
        if home_dir:
            cmd.extend(['--home', str(home_dir)])
        if system_user or password is None:
            cmd.append('--system')
        else:
            cmd.extend([
                '--create-home',
                '--shell', shell,
                '--password', password,
            ])
        if not primary_group:
            try:
                grp.getgrnam(username)
                primary_group = username  # avoid "group exists" error
            except KeyError:
                pass
        if primary_group:
            cmd.extend(['-g', primary_group])
        if secondary_groups:
            cmd.extend(['-G', ','.join(secondary_groups)])
        cmd.append(username)
        subprocess.check_call(cmd)
        user_info = pwd.getpwnam(username)
    return user_info


def user_exists(username):
    """Check if a user exists"""
    try:
        pwd.getpwnam(username)
        user_exists = True
    except KeyError:
        user_exists = False
    return user_exists


def uid_exists(uid):
    """Check if a uid exists"""
    try:
        pwd.getpwuid(uid)
        uid_exists = True
    except KeyError:
        uid_exists = False
    return uid_exists


def group_exists(groupname):
    """Check if a group exists"""
    try:
        grp.getgrnam(groupname)
        group_exists = True
    except KeyError:
        group_exists = False
    return group_exists


def gid_exists(gid):
    """Check if a gid exists"""
    try:
        grp.getgrgid(gid)
        gid_exists = True
    except KeyError:
        gid_exists = False
    return gid_exists


def add_group(group_name, system_group=False, gid=None):
    """Add a group to the system

    Will log but otherwise succeed if the group already exists.

    :param str group_name: group to create
    :param bool system_group: Create system group
    :param int gid: GID for user being created

    :returns: The password database entry struct, as returned by `grp.getgrnam`
    """
    try:
        group_info = grp.getgrnam(group_name)
        log('group {0} already exists!'.format(group_name))
        if gid:
            group_info = grp.getgrgid(gid)
            log('group with gid {0} already exists!'.format(gid))
    except KeyError:
        log('creating group {0}'.format(group_name))
        add_new_group(group_name, system_group, gid)
        group_info = grp.getgrnam(group_name)
    return group_info


def add_user_to_group(username, group):
    """Add a user to a group"""
    cmd = ['gpasswd', '-a', username, group]
    log("Adding user {} to group {}".format(username, group))
    subprocess.check_call(cmd)


def chage(username, lastday=None, expiredate=None, inactive=None,
          mindays=None, maxdays=None, root=None, warndays=None):
    """Change user password expiry information

    :param str username: User to update
    :param str lastday: Set when password was changed in YYYY-MM-DD format
    :param str expiredate: Set when user's account will no longer be
                           accessible in YYYY-MM-DD format.
                           -1 will remove an account expiration date.
    :param str inactive: Set the number of days of inactivity after a password
                         has expired before the account is locked.
                         -1 will remove an account's inactivity.
    :param str mindays: Set the minimum number of days between password
                        changes to MIN_DAYS.
                        0 indicates the password can be changed anytime.
    :param str maxdays: Set the maximum number of days during which a
                        password is valid.
                        -1 as MAX_DAYS will remove checking maxdays
    :param str root: Apply changes in the CHROOT_DIR directory
    :param str warndays: Set the number of days of warning before a password
                         change is required
    :raises subprocess.CalledProcessError: if call to chage fails
    """
    cmd = ['chage']
    if root:
        cmd.extend(['--root', root])
    if lastday:
        cmd.extend(['--lastday', lastday])
    if expiredate:
        cmd.extend(['--expiredate', expiredate])
    if inactive:
        cmd.extend(['--inactive', inactive])
    if mindays:
        cmd.extend(['--mindays', mindays])
    if maxdays:
        cmd.extend(['--maxdays', maxdays])
    if warndays:
        cmd.extend(['--warndays', warndays])
    cmd.append(username)
    subprocess.check_call(cmd)


remove_password_expiry = functools.partial(chage, expiredate='-1', inactive='-1', mindays='0', maxdays='-1')


def rsync(from_path, to_path, flags='-r', options=None, timeout=None):
    """Replicate the contents of a path"""
    options = options or ['--delete', '--executability']
    cmd = ['/usr/bin/rsync', flags]
    if timeout:
        cmd = ['timeout', str(timeout)] + cmd
    cmd.extend(options)
    cmd.append(from_path)
    cmd.append(to_path)
    log(" ".join(cmd))
    return subprocess.check_output(cmd, stderr=subprocess.STDOUT).decode('UTF-8').strip()


def symlink(source, destination):
    """Create a symbolic link"""
    log("Symlinking {} as {}".format(source, destination))
    cmd = [
        'ln',
        '-sf',
        source,
        destination,
    ]
    subprocess.check_call(cmd)


def mkdir(path, owner='root', group='root', perms=0o555, force=False):
    """Create a directory"""
    log("Making dir {} {}:{} {:o}".format(path, owner, group,
                                          perms))
    uid = pwd.getpwnam(owner).pw_uid
    gid = grp.getgrnam(group).gr_gid
    realpath = os.path.abspath(path)
    path_exists = os.path.exists(realpath)
    if path_exists and force:
        if not os.path.isdir(realpath):
            log("Removing non-directory file {} prior to mkdir()".format(path))
            os.unlink(realpath)
            os.makedirs(realpath, perms)
    elif not path_exists:
        os.makedirs(realpath, perms)
    os.chown(realpath, uid, gid)
    os.chmod(realpath, perms)


def write_file(path, content, owner='root', group='root', perms=0o444):
    """Create or overwrite a file with the contents of a byte string."""
    uid = pwd.getpwnam(owner).pw_uid
    gid = grp.getgrnam(group).gr_gid
    # lets see if we can grab the file and compare the context, to avoid doing
    # a write.
    existing_content = None
    existing_uid, existing_gid, existing_perms = None, None, None
    try:
        with open(path, 'rb') as target:
            existing_content = target.read()
        stat = os.stat(path)
        existing_uid, existing_gid, existing_perms = (
            stat.st_uid, stat.st_gid, stat.st_mode
        )
    except Exception:
        pass
    if content != existing_content:
        log("Writing file {} {}:{} {:o}".format(path, owner, group, perms),
            level=DEBUG)
        with open(path, 'wb') as target:
            os.fchown(target.fileno(), uid, gid)
            os.fchmod(target.fileno(), perms)
            if six.PY3 and isinstance(content, six.string_types):
                content = content.encode('UTF-8')
            target.write(content)
        return
    # the contents were the same, but we might still need to change the
    # ownership or permissions.
    if existing_uid != uid:
        log("Changing uid on already existing content: {} -> {}"
            .format(existing_uid, uid), level=DEBUG)
        os.chown(path, uid, -1)
    if existing_gid != gid:
        log("Changing gid on already existing content: {} -> {}"
            .format(existing_gid, gid), level=DEBUG)
        os.chown(path, -1, gid)
    if existing_perms != perms:
        log("Changing permissions on existing content: {} -> {}"
            .format(existing_perms, perms), level=DEBUG)
        os.chmod(path, perms)


def fstab_remove(mp):
    """Remove the given mountpoint entry from /etc/fstab"""
    return Fstab.remove_by_mountpoint(mp)


def fstab_add(dev, mp, fs, options=None):
    """Adds the given device entry to the /etc/fstab file"""
    return Fstab.add(dev, mp, fs, options=options)


def mount(device, mountpoint, options=None, persist=False, filesystem="ext3"):
    """Mount a filesystem at a particular mountpoint"""
    cmd_args = ['mount']
    if options is not None:
        cmd_args.extend(['-o', options])
    cmd_args.extend([device, mountpoint])
    try:
        subprocess.check_output(cmd_args)
    except subprocess.CalledProcessError as e:
        log('Error mounting {} at {}\n{}'.format(device, mountpoint, e.output))
        return False

    if persist:
        return fstab_add(device, mountpoint, filesystem, options=options)
    return True


def umount(mountpoint, persist=False):
    """Unmount a filesystem"""
    cmd_args = ['umount', mountpoint]
    try:
        subprocess.check_output(cmd_args)
    except subprocess.CalledProcessError as e:
        log('Error unmounting {}\n{}'.format(mountpoint, e.output))
        return False

    if persist:
        return fstab_remove(mountpoint)
    return True


def mounts():
    """Get a list of all mounted volumes as [[mountpoint,device],[...]]"""
    with open('/proc/mounts') as f:
        # [['/mount/point','/dev/path'],[...]]
        system_mounts = [m[1::-1] for m in [l.strip().split()
                                            for l in f.readlines()]]
    return system_mounts


def fstab_mount(mountpoint):
    """Mount filesystem using fstab"""
    cmd_args = ['mount', mountpoint]
    try:
        subprocess.check_output(cmd_args)
    except subprocess.CalledProcessError as e:
        log('Error unmounting {}\n{}'.format(mountpoint, e.output))
        return False
    return True


def file_hash(path, hash_type='md5'):
    """Generate a hash checksum of the contents of 'path' or None if not found.

    :param str hash_type: Any hash alrgorithm supported by :mod:`hashlib`,
                          such as md5, sha1, sha256, sha512, etc.
    """
    if os.path.exists(path):
        h = getattr(hashlib, hash_type)()
        with open(path, 'rb') as source:
            h.update(source.read())
        return h.hexdigest()
    else:
        return None


def path_hash(path):
    """Generate a hash checksum of all files matching 'path'. Standard
    wildcards like '*' and '?' are supported, see documentation for the 'glob'
    module for more information.

    :return: dict: A { filename: hash } dictionary for all matched files.
                   Empty if none found.
    """
    return {
        filename: file_hash(filename)
        for filename in glob.iglob(path)
    }


def check_hash(path, checksum, hash_type='md5'):
    """Validate a file using a cryptographic checksum.

    :param str checksum: Value of the checksum used to validate the file.
    :param str hash_type: Hash algorithm used to generate `checksum`.
        Can be any hash algorithm supported by :mod:`hashlib`,
        such as md5, sha1, sha256, sha512, etc.
    :raises ChecksumError: If the file fails the checksum

    """
    actual_checksum = file_hash(path, hash_type)
    if checksum != actual_checksum:
        raise ChecksumError("'%s' != '%s'" % (checksum, actual_checksum))


class ChecksumError(ValueError):
    """A class derived from Value error to indicate the checksum failed."""
    pass


class restart_on_change(object):
    """Decorator and context manager to handle restarts.

    Usage:

       @restart_on_change(restart_map, ...)
       def function_that_might_trigger_a_restart(...)
           ...

    Or:

       with restart_on_change(restart_map, ...):
           do_stuff_that_might_trigger_a_restart()
           ...
    """

    def __init__(self, restart_map, stopstart=False, restart_functions=None,
                 can_restart_now_f=None, post_svc_restart_f=None,
                 pre_restarts_wait_f=None):
        """
        :param restart_map: {file: [service, ...]}
        :type restart_map: Dict[str, List[str,]]
        :param stopstart: whether to stop, start or restart a service
        :type stopstart: booleean
        :param restart_functions: nonstandard functions to use to restart
                                  services {svc: func, ...}
        :type restart_functions: Dict[str, Callable[[str], None]]
        :param can_restart_now_f: A function used to check if the restart is
                                  permitted.
        :type can_restart_now_f: Callable[[str, List[str]], boolean]
        :param post_svc_restart_f: A function run after a service has
                                   restarted.
        :type post_svc_restart_f: Callable[[str], None]
        :param pre_restarts_wait_f: A function called before any restarts.
        :type pre_restarts_wait_f: Callable[None, None]
        """
        self.restart_map = restart_map
        self.stopstart = stopstart
        self.restart_functions = restart_functions
        self.can_restart_now_f = can_restart_now_f
        self.post_svc_restart_f = post_svc_restart_f
        self.pre_restarts_wait_f = pre_restarts_wait_f

    def __call__(self, f):
        """Work like a decorator.

        Returns a wrapped function that performs the restart if triggered.

        :param f: The function that is being wrapped.
        :type f: Callable[[Any], Any]
        :returns: the wrapped function
        :rtype: Callable[[Any], Any]
        """
        @functools.wraps(f)
        def wrapped_f(*args, **kwargs):
            return restart_on_change_helper(
                (lambda: f(*args, **kwargs)),
                self.restart_map,
                stopstart=self.stopstart,
                restart_functions=self.restart_functions,
                can_restart_now_f=self.can_restart_now_f,
                post_svc_restart_f=self.post_svc_restart_f,
                pre_restarts_wait_f=self.pre_restarts_wait_f)
        return wrapped_f

    def __enter__(self):
        """Enter the runtime context related to this object. """
        self.checksums = _pre_restart_on_change_helper(self.restart_map)

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Exit the runtime context related to this object.

        The parameters describe the exception that caused the context to be
        exited. If the context was exited without an exception, all three
        arguments will be None.
        """
        if exc_type is None:
            _post_restart_on_change_helper(
                self.checksums,
                self.restart_map,
                stopstart=self.stopstart,
                restart_functions=self.restart_functions,
                can_restart_now_f=self.can_restart_now_f,
                post_svc_restart_f=self.post_svc_restart_f,
                pre_restarts_wait_f=self.pre_restarts_wait_f)
        # All is good, so return False; any exceptions will propagate.
        return False


def restart_on_change_helper(lambda_f, restart_map, stopstart=False,
                             restart_functions=None,
                             can_restart_now_f=None,
                             post_svc_restart_f=None,
                             pre_restarts_wait_f=None):
    """Helper function to perform the restart_on_change function.

    This is provided for decorators to restart services if files described
    in the restart_map have changed after an invocation of lambda_f().

    This functions allows for a number of helper functions to be passed.

    `restart_functions` is a map with a service as the key and the
    corresponding value being the function to call to restart the service. For
    example if `restart_functions={'some-service': my_restart_func}` then
    `my_restart_func` should a function which takes one argument which is the
    service name to be retstarted.

    `can_restart_now_f` is a function which checks that a restart is permitted.
    It should return a bool which indicates if a restart is allowed and should
    take a service name (str) and a list of changed files (List[str]) as
    arguments.

    `post_svc_restart_f` is a function which runs after a service has been
    restarted. It takes the service name that was restarted as an argument.

    `pre_restarts_wait_f` is a function which is called before any restarts
    occur. The use case for this is an application which wants to try and
    stagger restarts between units.

    :param lambda_f: function to call.
    :type lambda_f: Callable[[], ANY]
    :param restart_map: {file: [service, ...]}
    :type restart_map: Dict[str, List[str,]]
    :param stopstart: whether to stop, start or restart a service
    :type stopstart: booleean
    :param restart_functions: nonstandard functions to use to restart services
                              {svc: func, ...}
    :type restart_functions: Dict[str, Callable[[str], None]]
    :param can_restart_now_f: A function used to check if the restart is
                              permitted.
    :type can_restart_now_f: Callable[[str, List[str]], boolean]
    :param post_svc_restart_f: A function run after a service has
                               restarted.
    :type post_svc_restart_f: Callable[[str], None]
    :param pre_restarts_wait_f: A function called before any restarts.
    :type pre_restarts_wait_f: Callable[None, None]
    :returns: result of lambda_f()
    :rtype: ANY
    """
    checksums = _pre_restart_on_change_helper(restart_map)
    r = lambda_f()
    _post_restart_on_change_helper(checksums,
                                   restart_map,
                                   stopstart,
                                   restart_functions,
                                   can_restart_now_f,
                                   post_svc_restart_f,
                                   pre_restarts_wait_f)
    return r


def _pre_restart_on_change_helper(restart_map):
    """Take a snapshot of file hashes.

    :param restart_map: {file: [service, ...]}
    :type restart_map: Dict[str, List[str,]]
    :returns: Dictionary of file paths and the files checksum.
    :rtype: Dict[str, str]
    """
    return {path: path_hash(path) for path in restart_map}


def _post_restart_on_change_helper(checksums,
                                   restart_map,
                                   stopstart=False,
                                   restart_functions=None,
                                   can_restart_now_f=None,
                                   post_svc_restart_f=None,
                                   pre_restarts_wait_f=None):
    """Check whether files have changed.

    :param checksums: Dictionary of file paths and the files checksum.
    :type checksums: Dict[str, str]
    :param restart_map: {file: [service, ...]}
    :type restart_map: Dict[str, List[str,]]
    :param stopstart: whether to stop, start or restart a service
    :type stopstart: booleean
    :param restart_functions: nonstandard functions to use to restart services
                              {svc: func, ...}
    :type restart_functions: Dict[str, Callable[[str], None]]
    :param can_restart_now_f: A function used to check if the restart is
                              permitted.
    :type can_restart_now_f: Callable[[str, List[str]], boolean]
    :param post_svc_restart_f: A function run after a service has
                               restarted.
    :type post_svc_restart_f: Callable[[str], None]
    :param pre_restarts_wait_f: A function called before any restarts.
    :type pre_restarts_wait_f: Callable[None, None]
    """
    if restart_functions is None:
        restart_functions = {}
    changed_files = defaultdict(list)
    restarts = []
    # create a list of lists of the services to restart
    for path, services in restart_map.items():
        if path_hash(path) != checksums[path]:
            restarts.append(services)
            for svc in services:
                changed_files[svc].append(path)
    # create a flat list of ordered services without duplicates from lists
    services_list = list(OrderedDict.fromkeys(itertools.chain(*restarts)))
    if services_list:
        if pre_restarts_wait_f:
            pre_restarts_wait_f()
        actions = ('stop', 'start') if stopstart else ('restart',)
        for service_name in services_list:
            if can_restart_now_f:
                if not can_restart_now_f(service_name,
                                         changed_files[service_name]):
                    continue
            if service_name in restart_functions:
                restart_functions[service_name](service_name)
            else:
                for action in actions:
                    service(action, service_name)
            if post_svc_restart_f:
                post_svc_restart_f(service_name)


def pwgen(length=None):
    """Generate a random password."""
    if length is None:
        # A random length is ok to use a weak PRNG
        length = random.choice(range(35, 45))
    alphanumeric_chars = [
        l for l in (string.ascii_letters + string.digits)
        if l not in 'l0QD1vAEIOUaeiou']
    # Use a crypto-friendly PRNG (e.g. /dev/urandom) for making the
    # actual password
    random_generator = random.SystemRandom()
    random_chars = [
        random_generator.choice(alphanumeric_chars) for _ in range(length)]
    return(''.join(random_chars))


def is_phy_iface(interface):
    """Returns True if interface is not virtual, otherwise False."""
    if interface:
        sys_net = '/sys/class/net'
        if os.path.isdir(sys_net):
            for iface in glob.glob(os.path.join(sys_net, '*')):
                if '/virtual/' in os.path.realpath(iface):
                    continue

                if interface == os.path.basename(iface):
                    return True

    return False


def get_bond_master(interface):
    """Returns bond master if interface is bond slave otherwise None.

    NOTE: the provided interface is expected to be physical
    """
    if interface:
        iface_path = '/sys/class/net/%s' % (interface)
        if os.path.exists(iface_path):
            if '/virtual/' in os.path.realpath(iface_path):
                return None

            master = os.path.join(iface_path, 'master')
            if os.path.exists(master):
                master = os.path.realpath(master)
                # make sure it is a bond master
                if os.path.exists(os.path.join(master, 'bonding')):
                    return os.path.basename(master)

    return None


def list_nics(nic_type=None):
    """Return a list of nics of given type(s)"""
    if isinstance(nic_type, six.string_types):
        int_types = [nic_type]
    else:
        int_types = nic_type

    interfaces = []
    if nic_type:
        for int_type in int_types:
            cmd = ['ip', 'addr', 'show', 'label', int_type + '*']
            ip_output = subprocess.check_output(
                cmd).decode('UTF-8', errors='replace')
            ip_output = ip_output.split('\n')
            ip_output = (line for line in ip_output if line)
            for line in ip_output:
                if line.split()[1].startswith(int_type):
                    matched = re.search('.*: (' + int_type +
                                        r'[0-9]+\.[0-9]+)@.*', line)
                    if matched:
                        iface = matched.groups()[0]
                    else:
                        iface = line.split()[1].replace(":", "")

                    if iface not in interfaces:
                        interfaces.append(iface)
    else:
        cmd = ['ip', 'a']
        ip_output = subprocess.check_output(
            cmd).decode('UTF-8', errors='replace').split('\n')
        ip_output = (line.strip() for line in ip_output if line)

        key = re.compile(r'^[0-9]+:\s+(.+):')
        for line in ip_output:
            matched = re.search(key, line)
            if matched:
                iface = matched.group(1)
                iface = iface.partition("@")[0]
                if iface not in interfaces:
                    interfaces.append(iface)

    return interfaces


def set_nic_mtu(nic, mtu):
    """Set the Maximum Transmission Unit (MTU) on a network interface."""
    cmd = ['ip', 'link', 'set', nic, 'mtu', mtu]
    subprocess.check_call(cmd)


def get_nic_mtu(nic):
    """Return the Maximum Transmission Unit (MTU) for a network interface."""
    cmd = ['ip', 'addr', 'show', nic]
    ip_output = subprocess.check_output(
        cmd).decode('UTF-8', errors='replace').split('\n')
    mtu = ""
    for line in ip_output:
        words = line.split()
        if 'mtu' in words:
            mtu = words[words.index("mtu") + 1]
    return mtu


def get_nic_hwaddr(nic):
    """Return the Media Access Control (MAC) for a network interface."""
    cmd = ['ip', '-o', '-0', 'addr', 'show', nic]
    ip_output = subprocess.check_output(cmd).decode('UTF-8', errors='replace')
    hwaddr = ""
    words = ip_output.split()
    if 'link/ether' in words:
        hwaddr = words[words.index('link/ether') + 1]
    return hwaddr


@contextmanager
def chdir(directory):
    """Change the current working directory to a different directory for a code
    block and return the previous directory after the block exits. Useful to
    run commands from a specified directory.

    :param str directory: The directory path to change to for this context.
    """
    cur = os.getcwd()
    try:
        yield os.chdir(directory)
    finally:
        os.chdir(cur)


def chownr(path, owner, group, follow_links=True, chowntopdir=False):
    """Recursively change user and group ownership of files and directories
    in given path. Doesn't chown path itself by default, only its children.

    :param str path: The string path to start changing ownership.
    :param str owner: The owner string to use when looking up the uid.
    :param str group: The group string to use when looking up the gid.
    :param bool follow_links: Also follow and chown links if True
    :param bool chowntopdir: Also chown path itself if True
    """
    uid = pwd.getpwnam(owner).pw_uid
    gid = grp.getgrnam(group).gr_gid
    if follow_links:
        chown = os.chown
    else:
        chown = os.lchown

    if chowntopdir:
        broken_symlink = os.path.lexists(path) and not os.path.exists(path)
        if not broken_symlink:
            chown(path, uid, gid)
    for root, dirs, files in os.walk(path, followlinks=follow_links):
        for name in dirs + files:
            full = os.path.join(root, name)
            try:
                chown(full, uid, gid)
            except (IOError, OSError) as e:
                # Intended to ignore "file not found". Catching both to be
                # compatible with both Python 2.7 and 3.x.
                if e.errno == errno.ENOENT:
                    pass


def lchownr(path, owner, group):
    """Recursively change user and group ownership of files and directories
    in a given path, not following symbolic links. See the documentation for
    'os.lchown' for more information.

    :param str path: The string path to start changing ownership.
    :param str owner: The owner string to use when looking up the uid.
    :param str group: The group string to use when looking up the gid.
    """
    chownr(path, owner, group, follow_links=False)


def owner(path):
    """Returns a tuple containing the username & groupname owning the path.

    :param str path: the string path to retrieve the ownership
    :return tuple(str, str): A (username, groupname) tuple containing the
                             name of the user and group owning the path.
    :raises OSError: if the specified path does not exist
    """
    stat = os.stat(path)
    username = pwd.getpwuid(stat.st_uid)[0]
    groupname = grp.getgrgid(stat.st_gid)[0]
    return username, groupname


def get_total_ram():
    """The total amount of system RAM in bytes.

    This is what is reported by the OS, and may be overcommitted when
    there are multiple containers hosted on the same machine.
    """
    with open('/proc/meminfo', 'r') as f:
        for line in f.readlines():
            if line:
                key, value, unit = line.split()
                if key == 'MemTotal:':
                    assert unit == 'kB', 'Unknown unit'
                    return int(value) * 1024  # Classic, not KiB.
        raise NotImplementedError()


UPSTART_CONTAINER_TYPE = '/run/container_type'


def is_container():
    """Determine whether unit is running in a container

    @return: boolean indicating if unit is in a container
    """
    if init_is_systemd():
        # Detect using systemd-detect-virt
        return subprocess.call(['systemd-detect-virt',
                                '--container']) == 0
    else:
        # Detect using upstart container file marker
        return os.path.exists(UPSTART_CONTAINER_TYPE)


def add_to_updatedb_prunepath(path, updatedb_path=UPDATEDB_PATH):
    """Adds the specified path to the mlocate's udpatedb.conf PRUNEPATH list.

    This method has no effect if the path specified by updatedb_path does not
    exist or is not a file.

    @param path: string the path to add to the updatedb.conf PRUNEPATHS value
    @param updatedb_path: the path the updatedb.conf file
    """
    if not os.path.exists(updatedb_path) or os.path.isdir(updatedb_path):
        # If the updatedb.conf file doesn't exist then don't attempt to update
        # the file as the package providing mlocate may not be installed on
        # the local system
        return

    with open(updatedb_path, 'r+') as f_id:
        updatedb_text = f_id.read()
        output = updatedb(updatedb_text, path)
        f_id.seek(0)
        f_id.write(output)
        f_id.truncate()


def updatedb(updatedb_text, new_path):
    lines = [line for line in updatedb_text.split("\n")]
    for i, line in enumerate(lines):
        if line.startswith("PRUNEPATHS="):
            paths_line = line.split("=")[1].replace('"', '')
            paths = paths_line.split(" ")
            if new_path not in paths:
                paths.append(new_path)
                lines[i] = 'PRUNEPATHS="{}"'.format(' '.join(paths))
    output = "\n".join(lines)
    return output


def modulo_distribution(modulo=3, wait=30, non_zero_wait=False):
    """ Modulo distribution

    This helper uses the unit number, a modulo value and a constant wait time
    to produce a calculated wait time distribution. This is useful in large
    scale deployments to distribute load during an expensive operation such as
    service restarts.

    If you have 1000 nodes that need to restart 100 at a time 1 minute at a
    time:

      time.wait(modulo_distribution(modulo=100, wait=60))
      restart()

    If you need restarts to happen serially set modulo to the exact number of
    nodes and set a high constant wait time:

      time.wait(modulo_distribution(modulo=10, wait=120))
      restart()

    @param modulo: int The modulo number creates the group distribution
    @param wait: int The constant time wait value
    @param non_zero_wait: boolean Override unit % modulo == 0,
                          return modulo * wait. Used to avoid collisions with
                          leader nodes which are often given priority.
    @return: int Calculated time to wait for unit operation
    """
    unit_number = int(local_unit().split('/')[1])
    calculated_wait_time = (unit_number % modulo) * wait
    if non_zero_wait and calculated_wait_time == 0:
        return modulo * wait
    else:
        return calculated_wait_time


def ca_cert_absolute_path(basename_without_extension):
    """Returns absolute path to CA certificate.

    :param basename_without_extension: Filename without extension
    :type basename_without_extension: str
    :returns: Absolute full path
    :rtype: str
    """
    return '{}/{}.crt'.format(CA_CERT_DIR, basename_without_extension)


def install_ca_cert(ca_cert, name=None):
    """
    Install the given cert as a trusted CA.

    The ``name`` is the stem of the filename where the cert is written, and if
    not provided, it will default to ``juju-{charm_name}``.

    If the cert is empty or None, or is unchanged, nothing is done.
    """
    if not ca_cert:
        return
    if not isinstance(ca_cert, bytes):
        ca_cert = ca_cert.encode('utf8')
    if not name:
        name = 'juju-{}'.format(charm_name())
    cert_file = ca_cert_absolute_path(name)
    new_hash = hashlib.md5(ca_cert).hexdigest()
    if file_hash(cert_file) == new_hash:
        return
    log("Installing new CA cert at: {}".format(cert_file), level=INFO)
    write_file(cert_file, ca_cert)
    subprocess.check_call(['update-ca-certificates', '--fresh'])


def get_system_env(key, default=None):
    """Get data from system environment as represented in ``/etc/environment``.

    :param key: Key to look up
    :type key: str
    :param default: Value to return if key is not found
    :type default: any
    :returns: Value for key if found or contents of default parameter
    :rtype: any
    :raises: subprocess.CalledProcessError
    """
    env_file = '/etc/environment'
    # use the shell and env(1) to parse the global environments file.  This is
    # done to get the correct result even if the user has shell variable
    # substitutions or other shell logic in that file.
    output = subprocess.check_output(
        ['env', '-i', '/bin/bash', '-c',
         'set -a && source {} && env'.format(env_file)],
        universal_newlines=True)
    for k, v in (line.split('=', 1)
                 for line in output.splitlines() if '=' in line):
        if k == key:
            return v
    else:
        return default



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2014-2015 Canonical Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import os

__author__ = 'Jorge Niedbalski R. <jorge.niedbalski@canonical.com>'


class Fstab(io.FileIO):
    """This class extends file in order to implement a file reader/writer
    for file `/etc/fstab`
    """

    class Entry(object):
        """Entry class represents a non-comment line on the `/etc/fstab` file
        """
        def __init__(self, device, mountpoint, filesystem,
                     options, d=0, p=0):
            self.device = device
            self.mountpoint = mountpoint
            self.filesystem = filesystem

            if not options:
                options = "defaults"

            self.options = options
            self.d = int(d)
            self.p = int(p)

        def __eq__(self, o):
            return str(self) == str(o)

        def __str__(self):
            return "{} {} {} {} {} {}".format(self.device,
                                              self.mountpoint,
                                              self.filesystem,
                                              self.options,
                                              self.d,
                                              self.p)

    DEFAULT_PATH = os.path.join(os.path.sep, 'etc', 'fstab')

    def __init__(self, path=None):
        if path:
            self._path = path
        else:
            self._path = self.DEFAULT_PATH
        super(Fstab, self).__init__(self._path, 'rb+')

    def _hydrate_entry(self, line):
        # NOTE: use split with no arguments to split on any
        #       whitespace including tabs
        return Fstab.Entry(*filter(
            lambda x: x not in ('', None),
            line.strip("\n").split()))

    @property
    def entries(self):
        self.seek(0)
        for line in self.readlines():
            line = line.decode('us-ascii')
            try:
                if line.strip() and not line.strip().startswith("#"):
                    yield self._hydrate_entry(line)
            except ValueError:
                pass

    def get_entry_by_attr(self, attr, value):
        for entry in self.entries:
            e_attr = getattr(entry, attr)
            if e_attr == value:
                return entry
        return None

    def add_entry(self, entry):
        if self.get_entry_by_attr('device', entry.device):
            return False

        self.write((str(entry) + '\n').encode('us-ascii'))
        self.truncate()
        return entry

    def remove_entry(self, entry):
        self.seek(0)

        lines = [l.decode('us-ascii') for l in self.readlines()]

        found = False
        for index, line in enumerate(lines):
            if line.strip() and not line.strip().startswith("#"):
                if self._hydrate_entry(line) == entry:
                    found = True
                    break

        if not found:
            return False

        lines.remove(line)

        self.seek(0)
        self.write(''.join(lines).encode('us-ascii'))
        self.truncate()
        return True

    @classmethod
    def remove_by_mountpoint(cls, mountpoint, path=None):
        fstab = cls(path=path)
        entry = fstab.get_entry_by_attr('mountpoint', mountpoint)
        if entry:
            return fstab.remove_entry(entry)
        return False

    @classmethod
    def add(cls, device, mountpoint, filesystem, options=None, path=None):
        return cls(path=path).add_entry(Fstab.Entry(device,
                                                    mountpoint, filesystem,
                                                    options=options))
